# This tests string handling operations

ZEN = """
The Zen of Python
Beautiful is better than ugly.
Explicit is better than implicit.
Simple is better than complex.
Complex is better than complicated.
Flat is better than nested.
Sparse is better than dense.
Readability counts.
Special cases aren't special enough to break the rules.
Although practicality beats purity.
Errors should never pass silently.
Unless explicitly silenced.
In the face of ambiguity, refuse the temptation to guess.
There should be one-- and preferably only one --obvious way to do it.
Although that way may not be obvious at first unless you're Dutch.
Now is better than never.
Although never is often better than *right* now.
If the implementation is hard to explain, it's a bad idea.
If the implementation is easy to explain, it may be a good idea.
Namespaces are one honking great idea -- let's do more of those!
"""


def test(niter):
    counts = {}
    for _ in range(niter):
        x = ZEN.replace("\n", " ").split(" ")
        y = " ".join(x)
        for i in range(50):
            a = ZEN[i : i * 2]
            b = a + "hello world"
        for c in ZEN:
            i = ord(c)
            c = chr(i)
    return (x[0], a)


###########################################################################
# Benchmark interface

bm_params = {
    (32, 10): (2,),
    (50, 10): (3,),
    (100, 10): (6,),
    (500, 10): (30,),
    (1000, 10): (60,),
    (5000, 10): (300,),
}


def bm_setup(params):
    (niter,) = params
    state = None

    def run():
        nonlocal state
        state = test(niter)

    def result():
        return niter, state

    return run, result



import sys

print((2**64).to_bytes(9, "little"))
print((2**64).to_bytes(9, "big"))

b = bytes(range(20))

il = int.from_bytes(b, "little")
ib = int.from_bytes(b, "big")
print(il)
print(ib)
print(il.to_bytes(20, "little"))
print(ib.to_bytes(20, "big"))

# check padding comes out correctly
print(il.to_bytes(40, "little"))
print(ib.to_bytes(40, "big"))

# check that extra zero bytes don't change the internal int value
print(int.from_bytes(b + bytes(10), "little") == int.from_bytes(b, "little"))

# can't write to a zero-length bytes object
try:
    ib.to_bytes(0, "little")
except OverflowError:
    print("OverflowError")

# or one that it too short
try:
    ib.to_bytes(18, "big")
except OverflowError:
    print("OverflowError")

# negative representations

# MicroPython int.to_bytes() behaves as if signed=True for negative numbers
if "micropython" in repr(sys.implementation):

    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e)
else:
    # Implement MicroPython compatible behaviour for CPython
    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e, signed=i < 0)


print(to_bytes_compat(-ib, 20, "big"))
print(to_bytes_compat(ib * -ib, 40, "big"))

# case where an additional byte is needed for sign bit
ib = (2**64) - 1
print(ib.to_bytes(8, "little"))

ib *= -1

try:
    print(to_bytes_compat(ib, 8, "little"))
except OverflowError:
    print("OverflowError")

print(to_bytes_compat(ib, 9, "little"))
print(to_bytes_compat(ib, 9, "big"))



"""Test cases for the fnmatch module."""

import unittest

from fnmatch import fnmatch, fnmatchcase, translate, filter


class FnmatchTestCase(unittest.TestCase):
    def check_match(self, filename, pattern, should_match=1, fn=fnmatch):
        if should_match:
            self.assertTrue(
                fn(filename, pattern),
                "expected %r to match pattern %r" % (filename, pattern),
            )
        else:
            self.assertTrue(
                not fn(filename, pattern),
                "expected %r not to match pattern %r" % (filename, pattern),
            )

    def test_fnmatch(self):
        check = self.check_match
        check("abc", "abc")
        check("abc", "?*?")
        check("abc", "???*")
        check("abc", "*???")
        check("abc", "???")
        check("abc", "*")
        check("abc", "ab[cd]")
        check("abc", "ab[!de]")
        check("abc", "ab[de]", 0)
        check("a", "??", 0)
        check("a", "b", 0)

        # these test that '\' is handled correctly in character sets;
        # see SF bug #409651
        check("\\", r"[\]")
        check("a", r"[!\]")
        check("\\", r"[!\]", 0)

        # test that filenames with newlines in them are handled correctly.
        # http://bugs.python.org/issue6665
        check("foo\nbar", "foo*")
        check("foo\nbar\n", "foo*")
        check("\nfoo", "foo*", False)
        check("\n", "*")

    def _test_mix_bytes_str(self):
        self.assertRaises(TypeError, fnmatch, "test", b"*")
        self.assertRaises(TypeError, fnmatch, b"test", "*")
        self.assertRaises(TypeError, fnmatchcase, "test", b"*")
        self.assertRaises(TypeError, fnmatchcase, b"test", "*")

    def test_fnmatchcase(self):
        check = self.check_match
        check("AbC", "abc", 0, fnmatchcase)
        check("abc", "AbC", 0, fnmatchcase)

    @unittest.skip("unsupported on MicroPython")
    def test_bytes(self):
        self.check_match(b"test", b"te*")
        self.check_match(b"test\xff", b"te*\xff")
        self.check_match(b"foo\nbar", b"foo*")


class TranslateTestCase(unittest.TestCase):
    def test_translate(self):
        self.assertEqual(translate("*"), "(?ms).*\Z")
        self.assertEqual(translate("?"), "(?ms).\Z")
        self.assertEqual(translate("a?b*"), "(?ms)a.b.*\Z")
        self.assertEqual(translate("[abc]"), "(?ms)[abc]\Z")
        self.assertEqual(translate("[]]"), "(?ms)[]]\Z")
        self.assertEqual(translate("[!x]"), "(?ms)[^x]\Z")
        self.assertEqual(translate("[^x]"), "(?ms)[\\^x]\Z")
        self.assertEqual(translate("[x"), "(?ms)\\[x\Z")


class FilterTestCase(unittest.TestCase):
    def test_filter(self):
        self.assertEqual(filter(["a", "b"], "a"), ["a"])



# test f-string debug feature {x=}


def f():
    return 4


def g(_):
    return 5


def h():
    return 6


x, y = 1, 2
print(f"{x=}")
print(f"{x=:08x}")
print(f"a {x=} b {y} c")
print(f"a {x=:08x} b {y} c")

print(f'a {f() + g("foo") + h()=} b')
print(f'a {f() + g("foo") + h()=:08x} b')

print(f"a {1,=} b")
print(f"a {x,y,=} b")
print(f"a {x,1=} b")



import pyb
from pyb import I2C

# test we can correctly create by id
for bus in (-1, 0, 1):
    try:
        I2C(bus)
        print("I2C", bus)
    except ValueError:
        print("ValueError", bus)

i2c = I2C(1)

i2c.init(I2C.CONTROLLER, baudrate=400000)
print(i2c.scan())
i2c.deinit()



# APA102 driver for MicroPython on ESP8266
# MIT license; Copyright (c) 2016 Robert Foss, Daniel Busch

from esp import apa102_write
from neopixel import NeoPixel


class APA102(NeoPixel):
    ORDER = (0, 1, 2, 3)

    def __init__(self, clock_pin, data_pin, n, bpp=4):
        super().__init__(data_pin, n, bpp)
        self.clock_pin = clock_pin
        self.clock_pin.init(clock_pin.OUT)

    def write(self):
        apa102_write(self.clock_pin, self.pin, self.buf)



# Test dictionary union.
#
# The tests sort the resulting dictionaries for now, since map insertion
# order is not implemented in MicroPython.

try:
    {} | {}
except TypeError:
    print("SKIP")
    raise SystemExit


def print_sorted_dict(d):
    print(sorted(d.items()))


def test_union(a, b):
    print_sorted_dict(a | b)
    print_sorted_dict(b | a)
    a |= a
    print_sorted_dict(a)
    a |= b
    print_sorted_dict(a)


d = {}
e = {}
test_union(d, e)

d = {1: "apple"}
e = {1: "cheese"}
test_union(d, e)

d = {"spam": 1, "eggs": 2, "cheese": 3}
e = {"cheese": "cheddar", "aardvark": "Ethel"}
test_union(d, e)



# test construction of array.array from different objects

try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

# tuple, list
print(array('b', (1, 2)))
print(array('h', [1, 2]))

# raw copy from bytes, bytearray
print(array('h', b'22'))  # should be byteorder-neutral
print(array('h', bytearray(2)))
print(array('i', bytearray(4)))

# convert from other arrays
print(array('H', array('b', [1, 2])))
print(array('b', array('I', [1, 2])))



# Tests constants of the math module available only with MICROPY_PY_MATH_CONSTANTS.
try:
    import math
    from math import isnan

    math.tau
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

print(math.tau == 2.0 * math.pi)

print(math.inf == float("inf"))
print(-math.inf == -float("inf"))

print(isnan(math.nan))
print(isnan(-math.nan))



class E:
    def __repr__(self):
        return "E"

    def __eq__(self, other):
        print('E eq', other)
        return 123

class F:
    def __repr__(self):
        return "F"

    def __ne__(self, other):
        print('F ne', other)
        return -456

print(E() != F())
print(F() != E())

tests = (None, 0, 1, 'a')

for val in tests:
    print('==== testing', val)
    print(E() == val)
    print(val == E())
    print(E() != val)
    print(val != E())
    print(F() == val)
    print(val == F())
    print(F() != val)
    print(val != F())



# Test ssl.SSLContext.verify_mode attribute, with additional CERT_xxx constants.
# These aren't available in the axtls implementation.

try:
    import ssl
except ImportError:
    print("SKIP")
    raise SystemExit

if not hasattr(ssl, "OPENSSL_VERSION") and not hasattr(ssl, "MBEDTLS_VERSION"):
    print("SKIP")
    raise SystemExit

ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

# Test setting and getting verify_mode.
ctx.verify_mode = ssl.CERT_OPTIONAL
print(ctx.verify_mode == ssl.CERT_OPTIONAL)
ctx.verify_mode = ssl.CERT_REQUIRED
print(ctx.verify_mode == ssl.CERT_REQUIRED)



# use accelerometer to test i2c bus

import pyb
from pyb import I2C

if not hasattr(pyb, "Accel"):
    print("SKIP")
    raise SystemExit

accel_addr = 76

pyb.Accel()  # this will init the MMA for us

i2c = I2C(1, I2C.CONTROLLER, baudrate=400000)

print(i2c.scan())
print(i2c.is_ready(accel_addr))

print(i2c.mem_read(1, accel_addr, 7, timeout=500))
i2c.mem_write(0, accel_addr, 0, timeout=500)

i2c.send(7, addr=accel_addr)
i2c.recv(1, addr=accel_addr)



try:
    import math

    math.factorial
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


for fun in (math.factorial,):
    for x in range(-1, 30):
        try:
            print("%d" % fun(x))
        except ValueError as e:
            print("ValueError")



try:
    import random
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    random.randint
except AttributeError:
    print("SKIP")
    raise SystemExit

print("randrange")
for i in range(50):
    assert 0 <= random.randrange(4) < 4
    assert 2 <= random.randrange(2, 6) < 6
    assert -2 <= random.randrange(-2, 2) < 2
    assert random.randrange(1, 9, 2) in (1, 3, 5, 7)
    assert random.randrange(2, 1, -1) in (1, 2)

# empty range
try:
    random.randrange(0)
except ValueError:
    print("ValueError")

# empty range
try:
    random.randrange(2, 1)
except ValueError:
    print("ValueError")

# zero step
try:
    random.randrange(2, 1, 0)
except ValueError:
    print("ValueError")

# empty range
try:
    random.randrange(2, 1, 1)
except ValueError:
    print("ValueError")

print("randint")
for i in range(50):
    assert 0 <= random.randint(0, 4) <= 4
    assert 2 <= random.randint(2, 6) <= 6
    assert -2 <= random.randint(-2, 2) <= 2

# empty range
try:
    random.randint(2, 1)
except ValueError:
    print("ValueError")

print("choice")
lst = [1, 2, 5, 6]
for i in range(50):
    assert random.choice(lst) in lst

# empty sequence
try:
    random.choice([])
except IndexError:
    print("IndexError")



"""Tests using unittest.subtest as an example reference for how it is used"""

import unittest


def factorial(value: int) -> int:
    """Iterative factorial algorithm implementation"""
    result = 1

    for i in range(1, value + 1):
        result *= i

    return result


class Person:
    """Represents a person with a name, age, and who can make friends"""

    def __init__(self, name: str, age: int) -> None:
        self.name = name
        self.age = age
        self.friends = set()

    def __repr__(self) -> str:
        return f"Person({self.name})"

    def add_friend(self, friend):
        """Logs that this Person has made a new friend"""
        self.friends.add(friend)

    def has_friend(self, friend_candidate) -> bool:
        """Determines if this Person has the friend `friend_candidate`"""
        return friend_candidate in self.friends

    def is_oldest_friend(self) -> bool:
        """Determines whether this Person is the oldest out of themself and their friends"""
        return self.age > max(friend.age for friend in self.friends)


class TestSubtest(unittest.TestCase):
    """Examples/tests of unittest.subTest()"""

    def test_sorted(self) -> None:
        """Test that the selection sort function correctly sorts lists"""
        tests = [
            {
                "unsorted": [-68, 15, 52, -54, -64, 20, 2, 66, 33],
                "sorted": [-68, -64, -54, 2, 15, 20, 33, 52, 66],
                "correct": True,
            },
            {
                "unsorted": [-68, 15, 52, -54, -64, 20, 2, 66, 33],
                "sorted": [-68, -54, -64, 2, 15, 20, 33, 52, 66],
                "correct": False,
            },
            {
                "unsorted": [-68, 15, 52, 54, -64, 20, 2, 66, 33],
                "sorted": [-68, -64, -54, 2, 15, 20, 33, 52, 66],
                "correct": False,
            },
            {
                "unsorted": [],
                "sorted": [],
                "correct": True,
            },
            {
                "unsorted": [42],
                "sorted": [42],
                "correct": True,
            },
            {
                "unsorted": [42],
                "sorted": [],
                "correct": False,
            },
            {
                "unsorted": [],
                "sorted": [24],
                "correct": False,
            },
            {
                "unsorted": [43, 44],
                "sorted": [43, 44],
                "correct": True,
            },
            {
                "unsorted": [44, 43],
                "sorted": [43, 44],
                "correct": True,
            },
        ]

        for test in tests:
            with self.subTest():  # Subtests continue to be tested, even if an earlier one fails
                if test["correct"]:  # Tests that match what is expected
                    self.assertEqual(sorted(test["unsorted"]), test["sorted"])
                else:  # Tests that are meant to fail
                    with self.assertRaises(AssertionError):
                        self.assertEqual(sorted(test["unsorted"]), test["sorted"])

    def test_factorial(self) -> None:
        """Test that the factorial fuction correctly calculates factorials

        Makes use of `msg` argument in subtest method to clarify which subtests had an
        error in the results
        """
        tests = [
            {
                "operand": 0,
                "result": 1,
                "correct": True,
            },
            {
                "operand": 1,
                "result": 1,
                "correct": True,
            },
            {
                "operand": 1,
                "result": 0,
                "correct": False,
            },
            {
                "operand": 2,
                "result": 2,
                "correct": True,
            },
            {
                "operand": 3,
                "result": 6,
                "correct": True,
            },
            {
                "operand": 3,
                "result": -6,
                "correct": False,
            },
            {
                "operand": 4,
                "result": 24,
                "correct": True,
            },
            {
                "operand": 15,
                "result": 1_307_674_368_000,
                "correct": True,
            },
            {
                "operand": 15,
                "result": 1_307_674_368_001,
                "correct": False,
            },
            {
                "operand": 11,
                "result": 39_916_800,
                "correct": True,
            },
        ]

        for test in tests:
            with self.subTest(
                f"{test['operand']}!"
            ):  # Let's us know we were testing "x!" when we get an error
                if test["correct"]:
                    self.assertEqual(factorial(test["operand"]), test["result"])
                else:
                    with self.assertRaises(AssertionError):
                        self.assertEqual(factorial(test["operand"]), test["result"])

    def test_person(self) -> None:
        """Test the Person class and its friend-making ability

        Makes use of subtest's params to specify relevant data about the tests, which is
        helpful for debugging
        """
        # Create a friendship
        alice = Person("Alice", 22)
        bob = Person("Bob", 23)
        alice.add_friend(bob)

        # Test friendship init
        with self.subTest(
            "Alice should have Bob as a friend", name=alice.name, friends=bob.friends
        ):  # Params `name` and `friends` provide useful data for debugging purposes
            self.assertTrue(alice.has_friend(bob))

        with self.subTest(
            "Bob should not have Alice as a friend", name=bob.name, friends=bob.friends
        ):
            self.assertFalse(bob.has_friend(alice))

        # Friendship is not always  commutative, so Bob is not implicitly friends with Alice
        with self.subTest("Alice and Bob should not both be friends with eachother"):
            with self.assertRaises(AssertionError):
                self.assertTrue(bob.has_friend(alice) and alice.has_friend(bob))

        # Bob becomes friends with Alice
        bob.add_friend(alice)

        with self.subTest(
            "Bob should now have Alice as a friend", name=bob.name, friends=bob.friends
        ):
            self.assertTrue(bob.has_friend(alice))

        with self.subTest(
            "Bob should be the oldest of his friends",
            age=bob.age,
            friend_ages=[friend.age for friend in bob.friends],
        ):  # Different params can be used for different subtests in the same test
            self.assertTrue(bob.is_oldest_friend())


if __name__ == "__main__":
    unittest.main()



# test that math functions support user classes with __float__

try:
    import math
except ImportError:
    print("SKIP")
    raise SystemExit


class TestFloat:
    def __float__(self):
        return 1.0


print("%.5g" % math.exp(TestFloat()))



import binascii
import socket
import ssl


# This self-signed key/cert pair is randomly generated and to be used for
# testing/demonstration only.  You should always generate your own key/cert.
key = binascii.unhexlify(
    b"3082013b020100024100cc20643fd3d9c21a0acba4f48f61aadd675f52175a9dcf07fbef"
    b"610a6a6ba14abb891745cd18a1d4c056580d8ff1a639460f867013c8391cdc9f2e573b0f"
    b"872d0203010001024100bb17a54aeb3dd7ae4edec05e775ca9632cf02d29c2a089b563b0"
    b"d05cdf95aeca507de674553f28b4eadaca82d5549a86058f9996b07768686a5b02cb240d"
    b"d9f1022100f4a63f5549e817547dca97b5c658038e8593cb78c5aba3c4642cc4cd031d86"
    b"8f022100d598d870ffe4a34df8de57047a50b97b71f4d23e323f527837c9edae88c79483"
    b"02210098560c89a70385c36eb07fd7083235c4c1184e525d838aedf7128958bedfdbb102"
    b"2051c0dab7057a8176ca966f3feb81123d4974a733df0f958525f547dfd1c271f9022044"
    b"6c2cafad455a671a8cf398e642e1be3b18a3d3aec2e67a9478f83c964c4f1f"
)
cert = binascii.unhexlify(
    b"308201d53082017f020203e8300d06092a864886f70d01010505003075310b3009060355"
    b"0406130258583114301206035504080c0b54686550726f76696e63653110300e06035504"
    b"070c075468654369747931133011060355040a0c0a436f6d70616e7958595a3113301106"
    b"0355040b0c0a436f6d70616e7958595a3114301206035504030c0b546865486f73744e61"
    b"6d65301e170d3139313231383033333935355a170d3239313231353033333935355a3075"
    b"310b30090603550406130258583114301206035504080c0b54686550726f76696e636531"
    b"10300e06035504070c075468654369747931133011060355040a0c0a436f6d70616e7958"
    b"595a31133011060355040b0c0a436f6d70616e7958595a3114301206035504030c0b5468"
    b"65486f73744e616d65305c300d06092a864886f70d0101010500034b003048024100cc20"
    b"643fd3d9c21a0acba4f48f61aadd675f52175a9dcf07fbef610a6a6ba14abb891745cd18"
    b"a1d4c056580d8ff1a639460f867013c8391cdc9f2e573b0f872d0203010001300d06092a"
    b"864886f70d0101050500034100b0513fe2829e9ecbe55b6dd14c0ede7502bde5d46153c8"
    b"e960ae3ebc247371b525caeb41bbcf34686015a44c50d226e66aef0a97a63874ca5944ef"
    b"979b57f0b3"
)


CONTENT = b"""\
HTTP/1.0 200 OK

Hello #%d from MicroPython!
"""


def main(use_stream=True):
    s = socket.socket()

    # Binding to all interfaces - server will be accessible to other hosts!
    ai = socket.getaddrinfo("0.0.0.0", 8443)
    print("Bind address info:", ai)
    addr = ai[0][-1]

    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(addr)
    s.listen(5)
    print("Listening, connect your browser to https://<this_host>:8443/")

    counter = 0
    while True:
        res = s.accept()
        client_s = res[0]
        client_addr = res[1]
        print("Client address:", client_addr)
        print("Client socket:", client_s)
        # CPython uses key keyfile/certfile arguments, but MicroPython uses key/cert
        client_s = ssl.wrap_socket(client_s, server_side=True, key=key, cert=cert)
        print(client_s)
        print("Request:")
        if use_stream:
            # Both CPython and MicroPython SSLSocket objects support read() and
            # write() methods.
            # Browsers are prone to terminate SSL connection abruptly if they
            # see unknown certificate, etc. We must continue in such case -
            # next request they issue will likely be more well-behaving and
            # will succeed.
            try:
                req = client_s.readline()
                print(req)
                while True:
                    h = client_s.readline()
                    if h == b"" or h == b"\r\n":
                        break
                    print(h)
                if req:
                    client_s.write(CONTENT % counter)
            except Exception as e:
                print("Exception serving request:", e)
        else:
            print(client_s.recv(4096))
            client_s.send(CONTENT % counter)
        client_s.close()
        counter += 1
        print()


main()



# test native function attributes


def f():
    pass


if not hasattr(f, "__name__"):
    print("SKIP")
    raise SystemExit


@micropython.native
def native_f():
    pass


print(type(native_f.__name__))
print(type(native_f.__globals__))
print(native_f.__globals__ is globals())

try:
    native_f.__name__ = None
except AttributeError:
    print("AttributeError")



from machine import Pin, Signal

# Red LED on pin LED_RED also known as A13
LED = Signal("LED_RED", Pin.OUT)

# Green LED on pin LED_GREEN also known as A14
LED2 = Signal("LED_GREEN", Pin.OUT)

# Yellow LED on pin LED_YELLOW also known as A15
LED3 = Signal("LED_YELLOW", Pin.OUT)

# Blue LED on pin LED_BLUE also known as B4
LED4 = Signal("LED_BLUE", Pin.OUT)



# test general errors with uctypes

try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

data = bytearray(b"01234567")

# first argument isn't an integer
try:
    uctypes.struct(data, {})
except TypeError:
    print("TypeError")

# del subscr not supported
S = uctypes.struct(uctypes.addressof(data), {})
try:
    del S[0]
except TypeError:
    print("TypeError")

# list is an invalid descriptor
S = uctypes.struct(uctypes.addressof(data), [])
try:
    S.x
except TypeError:
    print("TypeError")

# can't access attribute with invalid descriptor
S = uctypes.struct(uctypes.addressof(data), {"x": []})
try:
    S.x
except TypeError:
    print("TypeError")

# can't assign to aggregate
S = uctypes.struct(uctypes.addressof(data), {"x": (uctypes.ARRAY | 0, uctypes.INT8 | 2)})
try:
    S.x = 1
except TypeError:
    print("TypeError")

# unsupported unary op
try:
    hash(S)
except TypeError:
    print("TypeError")



try:
    from esp32 import Partition as p
    import micropython
except ImportError:
    print("SKIP")
    raise SystemExit

# try some vanilla OSError to get std error code
try:
    open("this filedoesnotexist", "r")
    print("FAILED TO RAISE")
except OSError as e:
    print(e)

# try to make nvs partition bootable, which ain't gonna work
part = p.find(type=p.TYPE_DATA)[0]
fun = p.set_boot
try:
    fun(part)
    print("FAILED TO RAISE")
except OSError as e:
    print(e)

# same but with out of memory condition by locking the heap
exc = "FAILED TO RAISE"
micropython.heap_lock()
try:
    fun(part)
except OSError as e:
    exc = e
micropython.heap_unlock()
print("exc:", exc)  # exc empty due to no memory

# same again but having an emergency buffer
micropython.alloc_emergency_exception_buf(256)
exc = "FAILED TO RAISE"
micropython.heap_lock()
try:
    fun(part)
except Exception as e:
    exc = e
micropython.heap_unlock()
print(exc)



try:
    from collections import deque
except ImportError:
    print("SKIP")
    raise SystemExit

# Initial sequence is supported
d = deque([1, 2, 3], 10)

# iteration over sequence
for x in d:
    print(x)

# Iterables larger than maxlen have the beginnings removed, also tests
# iteration through conversion to list
d = deque([1, 2, 3, 4, 5], 3)
print(list(d))

# Empty iterables are also supported
deque([], 10)

# Extending deques with iterables
d.extend([6, 7])
print(list(d))

# Accessing queue elements via index
d = deque((0, 1, 2, 3), 5)
print(d[0], d[1], d[-1])

# Writing queue elements via index
d[3] = 5
print(d[3])

# Accessing indices out of bounds raises IndexError
try:
    d[4]
except IndexError:
    print("IndexError")

try:
    d[4] = 0
except IndexError:
    print("IndexError")



freeze("$(PORT_DIR)/modules")
include("$(MPY_DIR)/extmod/asyncio")

# Useful networking-related packages.
require("bundle-networking")

# Require some micropython-lib modules.
require("aioespnow")
require("dht")
require("ds18x20")
require("neopixel")
require("onewire")
require("umqtt.robust")
require("umqtt.simple")
require("upysh")



import logging

logging.debug("test - debug")  # ignored by default
logging.info("test - info")  # ignored by default
logging.warning("test - warning")
logging.error("test - error")
logging.critical("test - critical")



import ffilib
import uctypes
import array
import os
from signal import *

libc = ffilib.libc()
librt = ffilib.open("librt")

CLOCK_REALTIME = 0
CLOCK_MONOTONIC = 1
SIGEV_SIGNAL = 0

sigval_t = {
    "sival_int": uctypes.INT32 | 0,
    "sival_ptr": (uctypes.PTR | 0, uctypes.UINT8),
}

sigevent_t = {
    "sigev_value": (0, sigval_t),
    "sigev_signo": uctypes.INT32 | 8,
    "sigev_notify": uctypes.INT32 | 12,
}

timespec_t = {
    "tv_sec": uctypes.INT32 | 0,
    "tv_nsec": uctypes.INT64 | 8,
}

itimerspec_t = {
    "it_interval": (0, timespec_t),
    "it_value": (16, timespec_t),
}


__libc_current_sigrtmin = libc.func("i", "__libc_current_sigrtmin", "")
SIGRTMIN = __libc_current_sigrtmin()

timer_create_ = librt.func("i", "timer_create", "ipp")
timer_settime_ = librt.func("i", "timer_settime", "PiPp")


def new(sdesc):
    buf = bytearray(uctypes.sizeof(sdesc))
    s = uctypes.struct(uctypes.addressof(buf), sdesc, uctypes.NATIVE)
    return s


def timer_create(sig_id):
    sev = new(sigevent_t)
    # print(sev)
    sev.sigev_notify = SIGEV_SIGNAL
    sev.sigev_signo = SIGRTMIN + sig_id
    timerid = array.array("P", [0])
    r = timer_create_(CLOCK_MONOTONIC, sev, timerid)
    os.check_error(r)
    # print("timerid", hex(timerid[0]))
    return timerid[0]


def timer_settime(tid, hz):
    period = 1000000000 // hz
    new_val = new(itimerspec_t)
    new_val.it_value.tv_nsec = period
    new_val.it_interval.tv_nsec = period
    # print("new_val:", bytes(new_val))
    old_val = new(itimerspec_t)
    # print(new_val, old_val)
    r = timer_settime_(tid, 0, new_val, old_val)
    os.check_error(r)
    # print("old_val:", bytes(old_val))
    # print("timer_settime", r)


class Timer:
    def __init__(self, id, freq):
        self.id = id
        self.tid = timer_create(id)
        self.freq = freq

    def callback(self, cb):
        self.cb = cb
        timer_settime(self.tid, self.freq)
        org_sig = signal(SIGRTMIN + self.id, self.handler)
        # print("Sig %d: %s" % (SIGRTMIN + self.id, org_sig))

    def handler(self, signum):
        # print('Signal handler called with signal', signum)
        self.cb(self)



# Test making HTTPS requests to sites that may require advanced ciphers.

import sys
import select
import socket
import ssl

# Don't run if ssl doesn't support required certificates (eg axtls).
if not hasattr(ssl, "CERT_REQUIRED"):
    print("SKIP")
    raise SystemExit


def test_one(site, opts):
    ai = socket.getaddrinfo(site, 443, socket.AF_INET)
    addr = ai[0][-1]

    s = socket.socket(socket.AF_INET)

    # Create SSLContext.
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    # CPython compatibility:
    # - disable check_hostname
    # - load default system certificate chain
    # - must wait for socket to be writable before calling wrap_socket
    if sys.implementation.name != "micropython":
        ssl_context.check_hostname = False
        ssl_context.load_default_certs()
        select.select([], [s], [])

    try:
        s.connect(addr)

        if "sni" in opts:
            s = ssl_context.wrap_socket(s, server_hostname=opts["host"])
        else:
            s = ssl_context.wrap_socket(s)

        s.write(b"GET / HTTP/1.0\r\nHost: %s\r\n\r\n" % bytes(site, "latin"))
        resp = s.read(4096)
        if resp[:7] != b"HTTP/1.":
            raise ValueError("response doesn't start with HTTP/1.")
        # print(resp)

    finally:
        s.close()


SITES = [
    "www.github.com",
    "micropython.org",
    "pypi.org",
    {"host": "api.pushbullet.com", "sni": True},
]


def main():
    for site in SITES:
        opts = {}
        if isinstance(site, dict):
            opts = site
            site = opts["host"]

        try:
            test_one(site, opts)
            print(site, "ok")
        except Exception as e:
            print(site, e)


main()



# Test struct pack/unpack with 'e' typecode.

try:
    import struct
except ImportError:
    print("SKIP")
    raise SystemExit

test_values = (
    1e-7,
    2e-7,
    1e-6,
    1e-5,
    1e-4,
    1e-3,
    1e-2,
    0.1,
    0,
    1,
    2,
    4,
    8,
    10,
    100,
    1e3,
    1e4,
    6e4,
    float("inf"),
)

for j in test_values:
    for i in (j, -j):
        x = struct.pack("<e", i)
        v = struct.unpack("<e", x)[0]
        print("%.7f %s %.15f %s" % (i, x, v, i == v))

# In CPython, packing a float that doesn't fit into a half-float raises OverflowError.
# But in MicroPython it does not, but rather stores the value as inf.
# This test is here for coverage.
try:
    struct.pack("e", 1e15)
except OverflowError:
    pass



# Driver test for official MicroPython LCD160CR display
# MIT license; Copyright (c) 2017 Damien P. George

import framebuf
import lcd160cr
import math
import time


def get_lcd(lcd):
    if isinstance(lcd, str):
        lcd = lcd160cr.LCD160CR(lcd)
    return lcd


def show_adc(lcd, adc):
    data = [adc.read_core_temp(), adc.read_core_vbat(), 3.3]
    try:
        data[2] = adc.read_vref()
    except:
        pass
    for i in range(3):
        lcd.set_text_color((825, 1625, 1600)[i], 0)
        if lcd.h == 160:
            lcd.set_font(2)
            lcd.set_pos(0, 100 + i * 16)
        else:
            lcd.set_font(2, trans=1)
            lcd.set_pos(0, lcd.h - 60 + i * 16)
        lcd.write("%4s: " % ("TEMP", "VBAT", "VREF")[i])
        if i > 0:
            s = "%6.3fV" % data[i]
        else:
            s = "%5.1f°C" % data[i]
        if lcd.h == 160:
            lcd.set_font(1, bold=0, scale=1)
        else:
            lcd.set_font(1, bold=0, scale=1, trans=1)
            lcd.set_pos(45, lcd.h - 60 + i * 16)
        lcd.write(s)


def test_features(lcd, orient=lcd160cr.PORTRAIT):
    # if we run on pyboard then use ADC and RTC features
    try:
        import pyb

        adc = pyb.ADCAll(12, 0xF0000)
        rtc = pyb.RTC()
    except:
        adc = None
        rtc = None

    # set orientation and clear screen
    lcd = get_lcd(lcd)
    lcd.set_orient(orient)
    lcd.set_pen(0, 0)
    lcd.erase()

    # create M-logo
    mlogo = framebuf.FrameBuffer(bytearray(17 * 17 * 2), 17, 17, framebuf.RGB565)
    mlogo.fill(0)
    mlogo.fill_rect(1, 1, 15, 15, 0xFFFFFF)
    mlogo.vline(4, 4, 12, 0)
    mlogo.vline(8, 1, 12, 0)
    mlogo.vline(12, 4, 12, 0)
    mlogo.vline(14, 13, 2, 0)

    # create inline framebuf
    offx = 14
    offy = 19
    w = 100
    h = 75
    fbuf = framebuf.FrameBuffer(bytearray(w * h * 2), w, h, framebuf.RGB565)
    lcd.set_spi_win(offx, offy, w, h)

    # initialise loop parameters
    tx = ty = 0
    t0 = time.ticks_us()

    for i in range(300):
        # update position of cross-hair
        t, tx2, ty2 = lcd.get_touch()
        if t:
            tx2 -= offx
            ty2 -= offy
            if tx2 >= 0 and ty2 >= 0 and tx2 < w and ty2 < h:
                tx, ty = tx2, ty2
        else:
            tx = (tx + 1) % w
            ty = (ty + 1) % h

        # create and show the inline framebuf
        fbuf.fill(lcd.rgb(128 + int(64 * math.cos(0.1 * i)), 128, 192))
        fbuf.line(
            w // 2,
            h // 2,
            w // 2 + int(40 * math.cos(0.2 * i)),
            h // 2 + int(40 * math.sin(0.2 * i)),
            lcd.rgb(128, 255, 64),
        )
        fbuf.hline(0, ty, w, lcd.rgb(64, 64, 64))
        fbuf.vline(tx, 0, h, lcd.rgb(64, 64, 64))
        fbuf.rect(tx - 3, ty - 3, 7, 7, lcd.rgb(64, 64, 64))
        for phase in (-0.2, 0, 0.2):
            x = w // 2 - 8 + int(50 * math.cos(0.05 * i + phase))
            y = h // 2 - 8 + int(32 * math.sin(0.05 * i + phase))
            fbuf.blit(mlogo, x, y)
        for j in range(-3, 3):
            fbuf.text(
                "MicroPython",
                5,
                h // 2 + 9 * j + int(20 * math.sin(0.1 * (i + j))),
                lcd.rgb(128 + 10 * j, 0, 128 - 10 * j),
            )
        lcd.show_framebuf(fbuf)

        # show results from the ADC
        if adc:
            show_adc(lcd, adc)

        # show the time
        if rtc:
            lcd.set_pos(2, 0)
            lcd.set_font(1)
            t = rtc.datetime()
            lcd.write(
                "%4d-%02d-%02d %2d:%02d:%02d.%01d"
                % (t[0], t[1], t[2], t[4], t[5], t[6], t[7] // 100000)
            )

        # compute the frame rate
        t1 = time.ticks_us()
        dt = time.ticks_diff(t1, t0)
        t0 = t1

        # show the frame rate
        lcd.set_pos(2, 9)
        lcd.write("%.2f fps" % (1000000 / dt))


def test_mandel(lcd, orient=lcd160cr.PORTRAIT):
    # set orientation and clear screen
    lcd = get_lcd(lcd)
    lcd.set_orient(orient)
    lcd.set_pen(0, 0xFFFF)
    lcd.erase()

    # function to compute Mandelbrot pixels
    def in_set(c):
        z = 0
        for i in range(32):
            z = z * z + c
            if abs(z) > 100:
                return i
        return 0

    # cache width and height of LCD
    w = lcd.w
    h = lcd.h

    # create the buffer for each line and set SPI parameters
    line = bytearray(w * 2)
    lcd.set_spi_win(0, 0, w, h)
    spi = lcd.fast_spi()

    # draw the Mandelbrot set line-by-line
    hh = (h - 1) / 3.2
    ww = (w - 1) / 2.4
    for v in range(h):
        for u in range(w):
            c = in_set((v / hh - 2.3) + (u / ww - 1.2) * 1j)
            if c < 16:
                rgb = c << 12 | c << 6
            else:
                rgb = 0xF800 | c << 6
            line[2 * u] = rgb
            line[2 * u + 1] = rgb >> 8
        spi.write(line)


def test_all(lcd, orient=lcd160cr.PORTRAIT):
    lcd = get_lcd(lcd)
    test_features(lcd, orient)
    test_mandel(lcd, orient)


print("To run all tests: test_all(<lcd>)")
print("Individual tests are: test_features, test_mandel")
print('<lcd> argument should be a connection, eg "X", or an LCD160CR object')



# test unicode in identifiers

# comment
# αβγδϵφζ

# global identifiers
α = 1
αβγ = 2
bβ = 3
βb = 4
print(α, αβγ, bβ, βb)


# function, argument, local identifiers
def α(β, γ):
    δ = β + γ
    print(β, γ, δ)


α(1, 2)


# class, method identifiers
class φ:
    def __init__(self):
        pass

    def δ(self, ϵ):
        print(ϵ)


zζzζz = φ()
if hasattr(zζzζz, "δ"):
    zζzζz.δ(ϵ=123)



import unittest


class TestUnittestAssertions(unittest.TestCase):
    def testFail(self):
        with self.assertRaises(AssertionError):
            self.fail("failure")

    def testEqual(self):
        self.assertEqual(0, 0)
        self.assertEqual([0, 1, 2], [0, 1, 2])
        with self.assertRaises(AssertionError):
            self.assertEqual(0, None)
        with self.assertRaises(AssertionError):
            self.assertEqual([0, 1, 2], [1, 2, 3])

    def test_AlmostEqual(self):
        self.assertAlmostEqual(1.00000001, 1.0)
        self.assertNotAlmostEqual(1.0000001, 1.0)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqual(1.0000001, 1.0)
        with self.assertRaises(AssertionError):
            self.assertNotAlmostEqual(1.00000001, 1.0)

        self.assertAlmostEqual(1.1, 1.0, places=0)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqual(1.1, 1.0, places=1)

        self.assertAlmostEqual(0, 0.1 + 0.1j, places=0)
        self.assertNotAlmostEqual(0, 0.1 + 0.1j, places=1)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqual(0, 0.1 + 0.1j, places=1)
        with self.assertRaises(AssertionError):
            self.assertNotAlmostEqual(0, 0.1 + 0.1j, places=0)

        self.assertAlmostEqual(float("inf"), float("inf"))
        with self.assertRaises(AssertionError):
            self.assertNotAlmostEqual(float("inf"), float("inf"))

    def test_AlmostEqualWithDelta(self):
        self.assertAlmostEqual(1.1, 1.0, delta=0.5)
        self.assertAlmostEqual(1.0, 1.1, delta=0.5)
        self.assertNotAlmostEqual(1.1, 1.0, delta=0.05)
        self.assertNotAlmostEqual(1.0, 1.1, delta=0.05)

        self.assertAlmostEqual(1.0, 1.0, delta=0.5)
        with self.assertRaises(AssertionError):
            self.assertNotAlmostEqual(1.0, 1.0, delta=0.5)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqual(1.1, 1.0, delta=0.05)
        with self.assertRaises(AssertionError):
            self.assertNotAlmostEqual(1.1, 1.0, delta=0.5)
        with self.assertRaises(TypeError):
            self.assertAlmostEqual(1.1, 1.0, places=2, delta=2)
        with self.assertRaises(TypeError):
            self.assertNotAlmostEqual(1.1, 1.0, places=2, delta=2)

    def testNotEqual(self):
        self.assertNotEqual([0, 1, 2], [0, 2, 1])
        with self.assertRaises(AssertionError):
            self.assertNotEqual(0, 0)
        with self.assertRaises(AssertionError):
            self.assertNotEqual([0, 1, 2], [0, 1, 2])

    def testIs(self):
        self.assertIs(None, None)
        with self.assertRaises(AssertionError):
            self.assertIs([1, 2, 3], [1, 2, 3])

    def testIsNot(self):
        self.assertIsNot([1, 2, 3], [1, 2, 3])
        with self.assertRaises(AssertionError):
            self.assertIsNot(None, None)

    def testIsNone(self):
        self.assertIsNone(None)
        with self.assertRaises(AssertionError):
            self.assertIsNone(0)

    def testIsNotNone(self):
        self.assertIsNotNone(0)
        with self.assertRaises(AssertionError):
            self.assertIsNotNone(None)

    def testTrue(self):
        self.assertTrue(True)
        with self.assertRaises(AssertionError):
            self.assertTrue(False)

    def testFalse(self):
        self.assertFalse(False)
        with self.assertRaises(AssertionError):
            self.assertFalse(True)

    def testIn(self):
        self.assertIn("t", "cat")
        with self.assertRaises(AssertionError):
            self.assertIn("x", "cat")

    def testIsInstance(self):
        self.assertIsInstance("cat", str)
        with self.assertRaises(AssertionError):
            self.assertIsInstance(7, str)

    def testRaises(self):
        with self.assertRaises(ZeroDivisionError):
            1 / 0
            pass

    @unittest.skip("test of skipping")
    def testSkip(self):
        self.fail("this should be skipped")

    def testAssert(self):
        e1 = None
        try:

            def func_under_test(a):
                assert a > 10

            self.assertRaises(AssertionError, func_under_test, 20)
        except AssertionError as e:
            e1 = e

        if not e1 or "not raised" not in e1.args[0]:
            self.fail("Expected to catch lack of AssertionError from assert in func_under_test")

    @unittest.expectedFailure
    def testExpectedFailure(self):
        self.assertEqual(1, 0)

    def testExpectedFailureNot(self):
        @unittest.expectedFailure
        def testInner():
            self.assertEqual(1, 1)

        try:
            testInner()
        except:
            pass
        else:
            self.fail("Unexpected success was not detected")

    def test_subtest_even(self):
        """
        Test that numbers between 0 and 5 are all even.
        """
        for i in range(0, 10, 2):
            with self.subTest("Should only pass for even numbers", i=i):
                self.assertEqual(i % 2, 0)


if __name__ == "__main__":
    unittest.main()



# MicroPython USB Mouse example
#
# To run this example:
#
# 1. Make sure `usb-device-mouse` is installed via: mpremote mip install usb-device-mouse
#
# 2. Run the example via: mpremote run mouse_example.py
#
# 3. mpremote will exit with an error after the previous step, because when the
#    example runs the existing USB device disconnects and then re-enumerates with
#    the mouse interface present. At this point, the example is running.
#
# 4. You should see the mouse move and right click. At this point, the example
#    is finished executing.
#
# To implement a more complex mouse with more buttons or other custom interface
# features, copy the usb-device-mouse/usb/device/mouse.py file into your own
# project and modify MouseInterface.
#
# MIT license; Copyright (c) 2023-2024 Angus Gratton
import time
import usb.device
from usb.device.mouse import MouseInterface


def mouse_example():
    m = MouseInterface()

    # Note: builtin_driver=True means that if there's a USB-CDC REPL
    # available then it will appear as well as the HID device.
    usb.device.get().init(m, builtin_driver=True)

    # wait for host to enumerate as a HID device...
    while not m.is_open():
        time.sleep_ms(100)

    time.sleep_ms(2000)

    print("Moving...")
    m.move_by(-100, 0)
    m.move_by(-100, 0)
    time.sleep_ms(500)

    print("Clicking...")
    m.click_right(True)
    time.sleep_ms(200)
    m.click_right(False)

    print("Done!")


mouse_example()



from micropython import const
from ubluepy import Peripheral, constants

BLE_GAP_ADV_FLAG_LE_GENERAL_DISC_MODE = const(0x02)
BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED = const(0x04)

BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE = const(
    BLE_GAP_ADV_FLAG_LE_GENERAL_DISC_MODE | BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED
)

EDDYSTONE_FRAME_TYPE_URL = const(0x10)
EDDYSTONE_URL_PREFIX_HTTP_WWW = const(0x00)  # "http://www".
EDDYSTONE_URL_SUFFIX_DOT_COM = const(0x01)  # ".com"


def string_to_binarray(text):
    b = bytearray([])
    for c in text:
        b.append(ord(c))
    return b


def gen_ad_type_content(ad_type, data):
    b = bytearray(1)
    b.append(ad_type)
    b.extend(data)
    b[0] = len(b) - 1
    return b


def generate_eddystone_adv_packet(url):
    # flags
    disc_mode = bytearray([BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE])
    packet_flags = gen_ad_type_content(constants.ad_types.AD_TYPE_FLAGS, disc_mode)

    # 16-bit uuid
    uuid = bytearray([0xAA, 0xFE])
    packet_uuid16 = gen_ad_type_content(
        constants.ad_types.AD_TYPE_16BIT_SERVICE_UUID_COMPLETE, uuid
    )

    # eddystone data
    rssi = 0xEE  # -18 dB, approx signal strength at 0m.
    eddystone_data = bytearray([])
    eddystone_data.append(EDDYSTONE_FRAME_TYPE_URL)
    eddystone_data.append(rssi)
    eddystone_data.append(EDDYSTONE_URL_PREFIX_HTTP_WWW)
    eddystone_data.extend(string_to_binarray(url))
    eddystone_data.append(EDDYSTONE_URL_SUFFIX_DOT_COM)

    # service data
    service_data = uuid + eddystone_data
    packet_service_data = gen_ad_type_content(
        constants.ad_types.AD_TYPE_SERVICE_DATA, service_data
    )

    # generate advertisement packet
    packet = bytearray([])
    packet.extend(packet_flags)
    packet.extend(packet_uuid16)
    packet.extend(packet_service_data)

    return packet


def start():
    adv_packet = generate_eddystone_adv_packet("micropython")
    p = Peripheral()
    p.advertise(data=adv_packet, connectable=False)



# test builtin pow() with integral values
# 3 arg version

try:
    print(pow(3, 4, 7))
except NotImplementedError:
    print("SKIP")
    raise SystemExit

# test some edge cases
print(pow(1, 1, 1))
print(pow(0, 1, 1))
print(pow(1, 0, 1))
print(pow(1, 0, 2))

# 3 arg pow is defined to only work on integers
try:
    print(pow("x", 5, 6))
except TypeError:
    print("TypeError expected")

try:
    print(pow(4, "y", 6))
except TypeError:
    print("TypeError expected")

try:
    print(pow(4, 5, "z"))
except TypeError:
    print("TypeError expected")



# Replace built-in collections module.
from ucollections import *

# Provide optional dependencies (which may be installed separately).
try:
    from .defaultdict import defaultdict
except ImportError:
    pass


class MutableMapping:
    pass



from urandom import *

_getrandbits32 = getrandbits


def getrandbits(bits: int) -> int:
    n = bits // 32
    d = 0
    for i in range(n):
        d |= _getrandbits32(32) << (i * 32)

    r = bits % 32
    if r >= 1:
        d |= _getrandbits32(r) << (n * 32)

    return d


def randrange(start, stop=None):
    if stop is None:
        stop = start
        start = 0
    upper = stop - start
    bits = 0
    pwr2 = 1
    while upper > pwr2:
        pwr2 <<= 1
        bits += 1
    while True:
        r = getrandbits(bits)
        if r < upper:
            break
    return r + start


def randint(start, stop):
    return randrange(start, stop + 1)


def shuffle(seq):
    l = len(seq)
    for i in range(l):
        j = randrange(l)
        seq[i], seq[j] = seq[j], seq[i]



# Test basic behaviour of asyncio.start_server()

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def test():
    # Test creating 2 servers using the same address
    print("create server1")
    server1 = await asyncio.start_server(None, "0.0.0.0", 8000)
    try:
        print("create server2")
        await asyncio.start_server(None, "0.0.0.0", 8000)
    except OSError as er:
        print("OSError")

    # Wait for server to close.
    async with server1:
        print("sleep")
        await asyncio.sleep(0)

    # Test that cancellation works before the server starts if
    # the subsequent code raises.
    print("create server3")
    server3 = await asyncio.start_server(None, "0.0.0.0", 8000)
    try:
        async with server3:
            raise OSError
    except OSError as er:
        print("OSError")

    # Test that closing doesn't raise CancelledError.
    print("create server4")
    server4 = await asyncio.start_server(None, "0.0.0.0", 8000)
    server4.close()
    await server4.wait_closed()
    print("server4 closed")

    # Test that cancelling the task will still raise CancelledError, checking
    # edge cases around how many times the tasks have been re-scheduled by
    # sleep.
    async def task(n):
        print("create task server", n)
        srv = await asyncio.start_server(None, "0.0.0.0", 8000)
        await srv.wait_closed()
        # This should be unreachable.
        print("task finished")

    for num_sleep in range(0, 5):
        print("sleep", num_sleep)
        t = asyncio.create_task(task(num_sleep))
        for _ in range(num_sleep):
            await asyncio.sleep(0)
        t.cancel()
        try:
            await t
        except asyncio.CancelledError:
            print("CancelledError")

    print("done")


asyncio.run(test())



# test that cmath functions support user classes with __float__ and __complex__

try:
    import cmath
except ImportError:
    print("SKIP")
    raise SystemExit


class TestFloat:
    def __float__(self):
        return 1.0


class TestComplex:
    def __complex__(self):
        return complex(10, 1)


for clas in TestFloat, TestComplex:
    print("%.5g" % cmath.phase(clas()))



# test builtin type

print(type(int))

try:
    type()
except TypeError:
    print('TypeError')

try:
    type(1, 2)
except TypeError:
    print('TypeError')

# second arg should be a tuple
try:
    type('abc', None, None)
except TypeError:
    print('TypeError')

# third arg should be a dict
try:
    type('abc', (), None)
except TypeError:
    print('TypeError')

# elements of second arg (the bases) should be types
try:
    type('abc', (1,), {})
except TypeError:
    print('TypeError')



import socket
import struct
from binascii import hexlify


class MQTTException(Exception):
    pass


class MQTTClient:
    def __init__(
        self,
        client_id,
        server,
        port=0,
        user=None,
        password=None,
        keepalive=0,
        ssl=None,
    ):
        if port == 0:
            port = 8883 if ssl else 1883
        self.client_id = client_id
        self.sock = None
        self.server = server
        self.port = port
        self.ssl = ssl
        self.pid = 0
        self.cb = None
        self.user = user
        self.pswd = password
        self.keepalive = keepalive
        self.lw_topic = None
        self.lw_msg = None
        self.lw_qos = 0
        self.lw_retain = False

    def _send_str(self, s):
        self.sock.write(struct.pack("!H", len(s)))
        self.sock.write(s)

    def _recv_len(self):
        n = 0
        sh = 0
        while 1:
            b = self.sock.read(1)[0]
            n |= (b & 0x7F) << sh
            if not b & 0x80:
                return n
            sh += 7

    def set_callback(self, f):
        self.cb = f

    def set_last_will(self, topic, msg, retain=False, qos=0):
        assert 0 <= qos <= 2
        assert topic
        self.lw_topic = topic
        self.lw_msg = msg
        self.lw_qos = qos
        self.lw_retain = retain

    def connect(self, clean_session=True, timeout=None):
        self.sock = socket.socket()
        self.sock.settimeout(timeout)
        addr = socket.getaddrinfo(self.server, self.port)[0][-1]
        self.sock.connect(addr)
        if self.ssl:
            self.sock = self.ssl.wrap_socket(self.sock, server_hostname=self.server)
        premsg = bytearray(b"\x10\0\0\0\0\0")
        msg = bytearray(b"\x04MQTT\x04\x02\0\0")

        sz = 10 + 2 + len(self.client_id)
        msg[6] = clean_session << 1
        if self.user:
            sz += 2 + len(self.user) + 2 + len(self.pswd)
            msg[6] |= 0xC0
        if self.keepalive:
            assert self.keepalive < 65536
            msg[7] |= self.keepalive >> 8
            msg[8] |= self.keepalive & 0x00FF
        if self.lw_topic:
            sz += 2 + len(self.lw_topic) + 2 + len(self.lw_msg)
            msg[6] |= 0x4 | (self.lw_qos & 0x1) << 3 | (self.lw_qos & 0x2) << 3
            msg[6] |= self.lw_retain << 5

        i = 1
        while sz > 0x7F:
            premsg[i] = (sz & 0x7F) | 0x80
            sz >>= 7
            i += 1
        premsg[i] = sz

        self.sock.write(premsg, i + 2)
        self.sock.write(msg)
        # print(hex(len(msg)), hexlify(msg, ":"))
        self._send_str(self.client_id)
        if self.lw_topic:
            self._send_str(self.lw_topic)
            self._send_str(self.lw_msg)
        if self.user:
            self._send_str(self.user)
            self._send_str(self.pswd)
        resp = self.sock.read(4)
        assert resp[0] == 0x20 and resp[1] == 0x02
        if resp[3] != 0:
            raise MQTTException(resp[3])
        return resp[2] & 1

    def disconnect(self):
        self.sock.write(b"\xe0\0")
        self.sock.close()

    def ping(self):
        self.sock.write(b"\xc0\0")

    def publish(self, topic, msg, retain=False, qos=0):
        pkt = bytearray(b"\x30\0\0\0")
        pkt[0] |= qos << 1 | retain
        sz = 2 + len(topic) + len(msg)
        if qos > 0:
            sz += 2
        assert sz < 2097152
        i = 1
        while sz > 0x7F:
            pkt[i] = (sz & 0x7F) | 0x80
            sz >>= 7
            i += 1
        pkt[i] = sz
        # print(hex(len(pkt)), hexlify(pkt, ":"))
        self.sock.write(pkt, i + 1)
        self._send_str(topic)
        if qos > 0:
            self.pid += 1
            pid = self.pid
            struct.pack_into("!H", pkt, 0, pid)
            self.sock.write(pkt, 2)
        self.sock.write(msg)
        if qos == 1:
            while 1:
                op = self.wait_msg()
                if op == 0x40:
                    sz = self.sock.read(1)
                    assert sz == b"\x02"
                    rcv_pid = self.sock.read(2)
                    rcv_pid = rcv_pid[0] << 8 | rcv_pid[1]
                    if pid == rcv_pid:
                        return
        elif qos == 2:
            assert 0

    def subscribe(self, topic, qos=0):
        assert self.cb is not None, "Subscribe callback is not set"
        pkt = bytearray(b"\x82\0\0\0")
        self.pid += 1
        struct.pack_into("!BH", pkt, 1, 2 + 2 + len(topic) + 1, self.pid)
        # print(hex(len(pkt)), hexlify(pkt, ":"))
        self.sock.write(pkt)
        self._send_str(topic)
        self.sock.write(qos.to_bytes(1, "little"))
        while 1:
            op = self.wait_msg()
            if op == 0x90:
                resp = self.sock.read(4)
                # print(resp)
                assert resp[1] == pkt[2] and resp[2] == pkt[3]
                if resp[3] == 0x80:
                    raise MQTTException(resp[3])
                return

    # Wait for a single incoming MQTT message and process it.
    # Subscribed messages are delivered to a callback previously
    # set by .set_callback() method. Other (internal) MQTT
    # messages processed internally.
    def wait_msg(self):
        res = self.sock.read(1)
        self.sock.setblocking(True)
        if res is None:
            return None
        if res == b"":
            raise OSError(-1)
        if res == b"\xd0":  # PINGRESP
            sz = self.sock.read(1)[0]
            assert sz == 0
            return None
        op = res[0]
        if op & 0xF0 != 0x30:
            return op
        sz = self._recv_len()
        topic_len = self.sock.read(2)
        topic_len = (topic_len[0] << 8) | topic_len[1]
        topic = self.sock.read(topic_len)
        sz -= topic_len + 2
        if op & 6:
            pid = self.sock.read(2)
            pid = pid[0] << 8 | pid[1]
            sz -= 2
        msg = self.sock.read(sz)
        self.cb(topic, msg)
        if op & 6 == 2:
            pkt = bytearray(b"\x40\x02\0\0")
            struct.pack_into("!H", pkt, 2, pid)
            self.sock.write(pkt)
        elif op & 6 == 4:
            assert 0
        return op

    # Checks whether a pending message from server is available.
    # If not, returns immediately with None. Otherwise, does
    # the same processing as wait_msg.
    def check_msg(self):
        self.sock.setblocking(False)
        return self.wait_msg()



#!/usr/bin/env python3
"""
Contract metadata handling for the NEAR Python contract compiler.
"""

import json
import tomllib
from pathlib import Path
from typing import Any, Dict

from .utils import console


def inject_metadata_function(contract_path: Path) -> Path:
    """
    Inject the contract_source_metadata function into a contract if it doesn't exist.

    Args:
        contract_path: Path to the contract file

    Returns:
        Path to the possibly modified contract file
    """
    # First check if the function already exists
    with open(contract_path) as f:
        content = f.read()

    if "def contract_source_metadata()" in content:
        return contract_path  # No injection needed

    # Initialize metadata with NEP-330 standard
    metadata = {"standards": [{"standard": "nep330", "version": "1.0.0"}]}

    # Extract metadata from pyproject.toml if it exists
    pyproject_path = contract_path.parent / "pyproject.toml"
    if pyproject_path.exists():
        try:
            metadata = extract_metadata_from_pyproject(pyproject_path, metadata)
        except Exception as e:
            console.print(
                f"[yellow]Warning: Could not read metadata from pyproject.toml: {e}"
            )

    # Get Git information for the contract
    from .reproducible import get_git_info

    git_info = get_git_info(contract_path.parent)

    # Add Git information directly to metadata
    if git_info:
        if "repository" in git_info and "commit" in git_info:
            if "build_info" not in metadata:
                metadata["build_info"] = {}  # type: ignore

            metadata["build_info"]["source_code_snapshot"] = (  # type: ignore
                f"git+{git_info['repository']}#{git_info['commit']}"
            )
    # Create the function code
    metadata_code = f"""

# Auto-generated NEP-330 metadata function
import near

@near.export
def contract_source_metadata():
    near.value_return('{json.dumps(metadata)}')
"""

    # Create a modified file with the appended function
    modified_path = contract_path.parent / f"{contract_path.stem}_with_metadata.py"
    with open(modified_path, "w") as f:
        f.write(content)
        f.write(metadata_code)

    console.print("[cyan]Added NEP-330 metadata function to contract[/]")

    return modified_path


def extract_metadata_from_pyproject(
    pyproject_path: Path, base_metadata: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Extract metadata from pyproject.toml using standard fields when possible.

    Args:
        pyproject_path: Path to the pyproject.toml file
        base_metadata: Initial metadata dictionary to extend

    Returns:
        Dict containing the updated metadata
    """
    with open(pyproject_path, "rb") as f:
        pyproject_data = tomllib.load(f)

    # Try to get metadata from project section (PEP 621 standard)
    project_data = pyproject_data.get("project", {})

    # Try to get NEAR-specific metadata from tool section
    near_data = pyproject_data.get("tool", {}).get("near", {}).get("contract", {})

    # Get reproducible build data if available
    reproducible_build = (
        pyproject_data.get("tool", {}).get("near", {}).get("reproducible_build", {})
    )

    # Map standard pyproject.toml fields to NEP-330 metadata
    if "version" in project_data:
        base_metadata["version"] = project_data["version"]
    elif "version" in near_data:
        base_metadata["version"] = near_data["version"]

    # Use standard URL field if available, fall back to NEAR-specific link
    if "urls" in project_data and "repository" in project_data["urls"]:
        base_metadata["link"] = project_data["urls"]["repository"]
    elif "url" in project_data:
        base_metadata["link"] = project_data["url"]
    elif "repository" in project_data:
        base_metadata["link"] = project_data["repository"]
    elif "link" in near_data:
        base_metadata["link"] = near_data["link"]

    # Use git repository if available from reproducible build info
    git_info = near_data.get("git_info", {})
    if git_info and "repository" in git_info and not base_metadata.get("link"):
        base_metadata["link"] = git_info["repository"]

    # Handle standards field
    if "standards" in near_data:
        standards = list(
            near_data["standards"]
        )  # Create a new list to avoid modifying the original

        # Ensure NEP-330 is included
        if not any(std.get("standard") == "nep330" for std in standards):
            standards.append({"standard": "nep330", "version": "1.0.0"})

        base_metadata["standards"] = standards

    # Handle build info
    if "build_info" in near_data:
        base_metadata["build_info"] = near_data["build_info"]

    # Add reproducible build information if available
    if reproducible_build:
        base_metadata["build_info"] = {
            "build_environment": reproducible_build.get("image", ""),
            "build_command": reproducible_build.get("container_build_command", []),
        }

        # Add git info if available
        if git_info:
            base_metadata["build_info"]["source_code_snapshot"] = (
                f"git+{git_info.get('repository', '')}#{git_info.get('commit', '')}"
            )

    return base_metadata



"""
The MIT License (MIT)

Copyright (c) 2023 Arduino SA
Copyright (c) 2018 KPN (Jan Bogaerts)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import io
import struct


class CBORDecodeError(Exception):
    """Raised when an error occurs deserializing a CBOR datastream."""


break_marker = object()


class CBORSimpleValue(object):
    """
    Represents a CBOR "simple value".
    :param int value: the value (0-255)
    """

    def __init__(self, value):
        if value < 0 or value > 255:
            raise TypeError("simple value too big")
        self.value = value

    def __eq__(self, other):
        if isinstance(other, CBORSimpleValue):
            return self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        return NotImplemented

    def __repr__(self):
        return "CBORSimpleValue({self.value})".format(self=self)


def decode_uint(decoder, subtype, allow_indefinite=False):
    # Major tag 0
    if subtype < 24:
        return subtype
    elif subtype == 24:
        return struct.unpack(">B", decoder.read(1))[0]
    elif subtype == 25:
        return struct.unpack(">H", decoder.read(2))[0]
    elif subtype == 26:
        return struct.unpack(">L", decoder.read(4))[0]
    elif subtype == 27:
        return struct.unpack(">Q", decoder.read(8))[0]
    elif subtype == 31 and allow_indefinite:
        return None
    else:
        raise CBORDecodeError("unknown unsigned integer subtype 0x%x" % subtype)


def decode_negint(decoder, subtype):
    # Major tag 1
    uint = decode_uint(decoder, subtype)
    return -uint - 1


def decode_bytestring(decoder, subtype):
    # Major tag 2
    length = decode_uint(decoder, subtype, allow_indefinite=True)
    if length is None:
        # Indefinite length
        buf = bytearray()
        while True:
            initial_byte = decoder.read(1)[0]
            if initial_byte == 255:
                return buf
            else:
                length = decode_uint(decoder, initial_byte & 31)
                value = decoder.read(length)
                buf.extend(value)
    else:
        return decoder.read(length)


def decode_string(decoder, subtype):
    # Major tag 3
    return decode_bytestring(decoder, subtype).decode("utf-8")


def decode_array(decoder, subtype):
    # Major tag 4
    items = []
    length = decode_uint(decoder, subtype, allow_indefinite=True)
    if length is None:
        # Indefinite length
        while True:
            value = decoder.decode()
            if value is break_marker:
                break
            else:
                items.append(value)
    else:
        for _ in range(length):
            item = decoder.decode()
            items.append(item)
    return items


def decode_map(decoder, subtype):
    # Major tag 5
    dictionary = {}
    length = decode_uint(decoder, subtype, allow_indefinite=True)
    if length is None:
        # Indefinite length
        while True:
            key = decoder.decode()
            if key is break_marker:
                break
            else:
                value = decoder.decode()
                dictionary[key] = value
    else:
        for _ in range(length):
            key = decoder.decode()
            value = decoder.decode()
            dictionary[key] = value

    return dictionary


def decode_special(decoder, subtype):
    # Simple value
    if subtype < 20:
        return CBORSimpleValue(subtype)

    # Major tag 7
    return special_decoders[subtype](decoder)


def decode_simple_value(decoder):
    return CBORSimpleValue(struct.unpack(">B", decoder.read(1))[0])


def decode_float16(decoder):
    decoder.read(2)
    raise NotImplementedError  # no float16 unpack function


def decode_float32(decoder):
    return struct.unpack(">f", decoder.read(4))[0]


def decode_float64(decoder):
    return struct.unpack(">d", decoder.read(8))[0]


major_decoders = {
    0: decode_uint,
    1: decode_negint,
    2: decode_bytestring,
    3: decode_string,
    4: decode_array,
    5: decode_map,
    7: decode_special,
}

special_decoders = {
    20: lambda self: False,
    21: lambda self: True,
    22: lambda self: None,
    # 23 is undefined
    24: decode_simple_value,
    25: decode_float16,
    26: decode_float32,
    27: decode_float64,
    31: lambda self: break_marker,
}


class CBORDecoder(object):
    """
    Deserializes a CBOR encoded byte stream.
    """

    def __init__(self, fp):
        self.fp = fp

    def read(self, amount):
        """
        Read bytes from the data stream.
        :param int amount: the number of bytes to read
        """
        data = self.fp.read(amount)
        if len(data) < amount:
            raise CBORDecodeError(
                "premature end of stream (expected to read {} bytes, got {} instead)".format(
                    amount, len(data)
                )
            )

        return data

    def decode(self):
        """
        Decode the next value from the stream.
        :raises CBORDecodeError: if there is any problem decoding the stream
        """
        try:
            initial_byte = self.fp.read(1)[0]
            major_type = initial_byte >> 5
            subtype = initial_byte & 31
        except Exception as e:
            raise CBORDecodeError(
                "error reading major type at index {}: {}".format(self.fp.tell(), e)
            )

        decoder = major_decoders[major_type]
        try:
            return decoder(self, subtype)
        except CBORDecodeError:
            raise
        except Exception as e:
            raise CBORDecodeError(
                "error decoding value {}".format(e)
            )  # tell doesn't work on micropython at the moment


def loads(payload, **kwargs):
    """
    Deserialize an object from a bytestring.
    :param bytes payload: the bytestring to serialize
    :param kwargs: keyword arguments passed to :class:`~.CBORDecoder`
    :return: the deserialized object
    """
    fp = io.BytesIO(payload)
    return CBORDecoder(fp, **kwargs).decode()


def load(fp, **kwargs):
    """
    Deserialize an object from an open file.
    :param fp: the input file (any file-like object)
    :param kwargs: keyword arguments passed to :class:`~.CBORDecoder`
    :return: the deserialized object
    """
    return CBORDecoder(fp, **kwargs).decode()



# test custom builtin module

try:
    import cexample
except ImportError:
    print("SKIP")
    raise SystemExit

print(cexample)
print(cexample.__name__)

d = dir(cexample)
d.index("add_ints")
d.index("Timer")
d.index("AdvancedTimer")

print(cexample.add_ints(1, 3))



# Implementation of a very simple, custom USB device in Python.  The device has an
# IN and OUT endpoint, accepts up to 64 bytes of data on the OUT endpoint, and echos
# that data back to the IN endpoint.
#
# To run, just execute this file on a device with machine.USBDevice support.  The device
# will then change to the custom USB device.
#
# For example, use `mpremote` (the `--no-follow` option starts the script running
# without waiting for a response, because there won't be a response after the device
# changes its USB mode):
#
# $ mpremote run --no-follow usb_simple_device.py
#
# Then, run the host side of the example on your PC using:
#
# $ sudo python usb_simple_host_pyusb.py
#
# You'll need to have `pyusb` installed via `pip install pyusb` to run the host PC code.
# And `sudo` is most likely needed to access the custom USB device.
#
# Once you have finished running the code, you will need to reset or unplug the USB
# device to stop it.

import machine

# VID and PID of the USB device.
VID = 0xF055
PID = 0x9999

# USB endpoints used by the device.
EP_OUT = 0x01
EP_IN = 0x81

# USB device descriptor.
_desc_dev = bytes(
    [
        0x12,  # bLength
        0x01,  # bDescriptorType: Device
        0x00,
        0x02,  # USB version: 2.00
        0xFF,  # bDeviceClass: vendor
        0x00,  # bDeviceSubClass
        0x01,  # bDeviceProtocol
        0x40,  # bMaxPacketSize
        VID & 0xFF,
        VID >> 8 & 0xFF,  # VID
        PID & 0xFF,
        PID >> 8 & 0xFF,  # PID
        0x00,
        0x01,  # bcdDevice: 1.00
        0x11,  # iManufacturer
        0x12,  # iProduct
        0x13,  # iSerialNumber
        0x01,  # bNumConfigurations: 1
    ]
)

# USB configuration descriptor.
_desc_cfg = bytes(
    [
        # Configuration Descriptor.
        0x09,  # bLength
        0x02,  # bDescriptorType: configuration
        0x20,
        0x00,  # wTotalLength: 32
        0x01,  # bNumInterfaces
        0x01,  # bConfigurationValue
        0x14,  # iConfiguration
        0xA0,  # bmAttributes
        0x96,  # bMaxPower
        # Interface Descriptor.
        0x09,  # bLength
        0x04,  # bDescriptorType: interface
        0x00,  # bInterfaceNumber
        0x00,  # bAlternateSetting
        0x02,  # bNumEndpoints
        0xFF,  # bInterfaceClass
        0x03,  # bInterfaceSubClass
        0x00,  # bInterfaceProtocol
        0x15,  # iInterface
        # Endpoint IN1.
        0x07,  # bLength
        0x05,  # bDescriptorType: endpoint
        EP_IN,  # bEndpointAddress
        0x03,  # bmAttributes: interrupt
        0x40,
        0x00,  # wMaxPacketSize
        0x0A,  # bInterval
        # Endpoint OUT1.
        0x07,  # bLength
        0x05,  # bDescriptorType: endpoint
        EP_OUT,  # bEndpointAddress
        0x02,  # bmAttributes: bulk
        0x40,
        0x00,  # wMaxPacketSize
        0x00,  # bInterval
    ]
)

# USB strings.
_desc_strs = {
    0x11: b"iManufacturer",
    0x12: b"iProduct",
    0x13: b"iSerial",
    0x14: b"iInterface",
    0x15: b"iInterface",
    0x16: b"Extra information",
}


# USB callback for when our custom USB interface is opened by the host.
def _open_itf_cb(interface_desc_view):
    print("_open_itf_cb", bytes(interface_desc_view))
    # Prepare to receive first data packet on the OUT endpoint.
    usbd.submit_xfer(EP_OUT, usbd_buf)


# USB callback for when a data transfer (IN or OUT) has completed.
def _xfer_cb(ep_addr, result, xferred_bytes):
    print("_xfer_cb", ep_addr, result, xferred_bytes)
    if ep_addr == EP_OUT:
        # Received data packet from the host, print it out.
        print(usbd_buf)
        # And then echo the data back to the host.
        usbd.submit_xfer(EP_IN, memoryview(usbd_buf)[:xferred_bytes])
    elif ep_addr == EP_IN:
        # Host got our data, prepare to receive the next data packet.
        usbd.submit_xfer(EP_OUT, usbd_buf)


# USB data buffer, for IN and OUT transfers.
usbd_buf = bytearray(64)

# Switch the USB device to our custom USB driver.
usbd = machine.USBDevice()
usbd.builtin_driver = usbd.BUILTIN_NONE
usbd.config(
    desc_dev=_desc_dev,
    desc_cfg=_desc_cfg,
    desc_strs=_desc_strs,
    open_itf_cb=_open_itf_cb,
    xfer_cb=_xfer_cb,
)
usbd.active(1)



# Test MicroPython extensions on CPython asyncio:
# - sleep_ms
# - wait_for_ms

try:
    import time, asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def task(id, t):
    print("task start", id)
    await asyncio.sleep_ms(t)
    print("task end", id)
    return id * 2


async def main():
    # Simple sleep_ms
    t0 = time.ticks_ms()
    await asyncio.sleep_ms(1)
    print(time.ticks_diff(time.ticks_ms(), t0) < 100)

    try:
        # Sleep 1ms beyond maximum allowed sleep value
        await asyncio.sleep_ms(time.ticks_add(0, -1) // 2 + 1)
    except OverflowError:
        print("OverflowError")

    # When task finished before the timeout
    print(await asyncio.wait_for_ms(task(1, 5), 50))

    # When timeout passes and task is cancelled
    try:
        print(await asyncio.wait_for_ms(task(2, 50), 5))
    except asyncio.TimeoutError:
        print("timeout")

    print("finish")


asyncio.run(main())



try:
    import struct
except ImportError:
    print("SKIP")
    raise SystemExit

print(struct.calcsize("<bI"))
print(struct.unpack("<bI", b"\x80\0\0\x01\0"))
print(struct.calcsize(">bI"))
print(struct.unpack(">bI", b"\x80\0\0\x01\0"))

# 32-bit little-endian specific
#print(struct.unpack("bI", b"\x80\xaa\x55\xaa\0\0\x01\0"))

print(struct.pack("<l", 1))
print(struct.pack(">l", 1))
print(struct.pack("<i", 1))
print(struct.pack(">i", 1))
print(struct.pack("<h", 1))
print(struct.pack(">h", 1))
print(struct.pack("<b", 1))
print(struct.pack(">b", 1))
print(struct.pack("<x"))
print(struct.pack(">x"))

print(struct.pack("<bI", -128, 256))
print(struct.pack(">bI", -128, 256))

print(struct.calcsize("100sI"))
print(struct.calcsize("97sI"))
print(struct.unpack("<6sH", b"foo\0\0\0\x12\x34"))
print(struct.pack("<6sH", b"foo", 10000))

print(struct.calcsize("7xx"))
print(struct.pack("7xx"))

print(struct.calcsize(">bxI3xH"))
print(struct.pack(">bxI3xH", 1, 2, 3))
print(struct.unpack(">bxI3xH", b"\x01\0\0\0\0\x02\0\0\0\0\x03"))

s = struct.pack("BHBI", 10, 100, 200, 300)
v = struct.unpack("BHBI", s)
print(v == (10, 100, 200, 300))

# network byte order
print(struct.pack('!i', 123))

# check that we get an error if the buffer is too small
try:
    struct.unpack('I', b'\x00\x00\x00')
except:
    print('struct.error')

# first arg must be a string
try:
    struct.pack(1, 2)
except TypeError:
    print('TypeError')

# make sure that unknown types are detected
try:
    struct.pack("z", 1)
except:
    print("Unknown type")

# Initially repetition counters were supported only for strings,
# but later were implemented for all.
print(struct.unpack("<3B2h", b"foo\x12\x34\xff\xff"))
print(struct.pack("<3B", 1, 2, 3))

# pack_into
buf = bytearray(b'>>>123<<<')
struct.pack_into('<bbb', buf, 3, 0x41, 0x42, 0x43)
print(buf)
struct.pack_into('<bbb', buf, -6, 0x44, 0x45, 0x46)
print(buf)

# check that we get an error if the buffer is too small
try:
    struct.pack_into('I', bytearray(1), 0, 0)
except:
    print('struct.error')

try:
    struct.pack_into('<bbb', buf, 7, 0x41, 0x42, 0x43)
except:
    print('struct.error')
try:
    struct.pack_into('<bbb', buf, -10, 0x41, 0x42, 0x43)
except:
    print('struct.error')

# unpack_from
buf = b'0123456789'
print(struct.unpack_from('<b', buf, 4))
print(struct.unpack_from('<b', buf, -4))
try:
    print(struct.unpack_from('<b', buf, 10))
except:
    print('struct.error')
try:
    print(struct.unpack_from('<b', buf, -11))
except:
    print('struct.error')



# common tests for ffi_int_types/long32/long64
# requires ffi_lib.c to be compiled as: $(CC) -shared -o ffi_lib.so ffi_lib.c

import os, sys

try:
    import ffi
except ImportError:
    print("SKIP")
    raise SystemExit

ffi_lib_filename = sys.argv[0].rsplit("/", 1)[0] + "/ffi_lib.so"
try:
    os.stat(ffi_lib_filename)
except OSError:
    print("SKIP")
    raise SystemExit

ffi_lib = ffi.open(ffi_lib_filename)


def test(funcs):
    for type, name in funcs:
        func = ffi_lib.func(type, name, type)
        for val in (
            0,
            0x7F,
            0x80,
            0xFF,
            0x100,
            0x7FFF,
            0x8000,
            0xFFFF,
            0x10000,
            0x7FFFFFFF,
            0x80000000,
            0xFFFFFFFF,
            0x100000000,
            0x7FFF_FFFF_FFFF_FFFF,
            0x8000_0000_0000_0000,
            0xFFFF_FFFF_FFFF_FFFF,
            0x1_0000_0000_0000_0000,
        ):
            print("{}({:x}) = {:x}".format(name, val, func(val)))


if __name__ == "__main__":
    print("SKIP")
    raise SystemExit



try:
    import re
except ImportError:
    print("SKIP")
    raise SystemExit

r = re.compile(" ")
s = r.split("a b c foobar")
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar")
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar", 1)
print(s)

r = re.compile(" +")
s = r.split("a b    c   foobar", 2)
print(s)

r = re.compile("[a-f]+")
s = r.split("0a3b9")
print(s)

# bytes objects
r = re.compile(b"x")
s = r.split(b"fooxbar")
print(s)

# using ^
r = re.compile("^ab")
s = r.split("abababcabab")
print(s)

# using ^ with |
r = re.compile("^ab|cab")
s = r.split("abababcabab")
print(s)



# MicroPython USB hid module
#
# This implements a base HIDInterface class that can be used directly,
# or subclassed into more specific HID interface types.
#
# MIT license; Copyright (c) 2023 Angus Gratton
from micropython import const
import machine
import struct
import time
from .core import Interface, Descriptor, split_bmRequestType

_EP_IN_FLAG = const(1 << 7)

# Control transfer stages
_STAGE_IDLE = const(0)
_STAGE_SETUP = const(1)
_STAGE_DATA = const(2)
_STAGE_ACK = const(3)

# Request types
_REQ_TYPE_STANDARD = const(0x0)
_REQ_TYPE_CLASS = const(0x1)
_REQ_TYPE_VENDOR = const(0x2)
_REQ_TYPE_RESERVED = const(0x3)

# Descriptor types
_DESC_HID_TYPE = const(0x21)
_DESC_REPORT_TYPE = const(0x22)
_DESC_PHYSICAL_TYPE = const(0x23)

# Interface and protocol identifiers
_INTERFACE_CLASS = const(0x03)
_INTERFACE_SUBCLASS_NONE = const(0x00)
_INTERFACE_SUBCLASS_BOOT = const(0x01)

_INTERFACE_PROTOCOL_NONE = const(0x00)
_INTERFACE_PROTOCOL_KEYBOARD = const(0x01)
_INTERFACE_PROTOCOL_MOUSE = const(0x02)

# bRequest values for HID control requests
_REQ_CONTROL_GET_REPORT = const(0x01)
_REQ_CONTROL_GET_IDLE = const(0x02)
_REQ_CONTROL_GET_PROTOCOL = const(0x03)
_REQ_CONTROL_GET_DESCRIPTOR = const(0x06)
_REQ_CONTROL_SET_REPORT = const(0x09)
_REQ_CONTROL_SET_IDLE = const(0x0A)
_REQ_CONTROL_SET_PROTOCOL = const(0x0B)

# Standard descriptor lengths
_STD_DESC_INTERFACE_LEN = const(9)
_STD_DESC_ENDPOINT_LEN = const(7)


class HIDInterface(Interface):
    # Abstract base class to implement a USB device HID interface in Python.

    def __init__(
        self,
        report_descriptor,
        extra_descriptors=[],
        set_report_buf=None,
        protocol=_INTERFACE_PROTOCOL_NONE,
        interface_str=None,
    ):
        # Construct a new HID interface.
        #
        # - report_descriptor is the only mandatory argument, which is the binary
        # data consisting of the HID Report Descriptor. See Device Class
        # Definition for Human Interface Devices (HID) v1.11 section 6.2.2 Report
        # Descriptor, p23.
        #
        # - extra_descriptors is an optional argument holding additional HID
        #   descriptors, to append after the mandatory report descriptor. Most
        #   HID devices do not use these.
        #
        # - set_report_buf is an optional writable buffer object (i.e.
        #   bytearray), where SET_REPORT requests from the host can be
        #   written. Only necessary if the report_descriptor contains Output
        #   entries. If set, the size must be at least the size of the largest
        #   Output entry.
        #
        # - protocol can be set to a specific value as per HID v1.11 section 4.3 Protocols, p9.
        #
        # - interface_str is an optional string descriptor to associate with the HID USB interface.
        super().__init__()
        self.report_descriptor = report_descriptor
        self.extra_descriptors = extra_descriptors
        self._set_report_buf = set_report_buf
        self.protocol = protocol
        self.interface_str = interface_str

        self._int_ep = None  # set during enumeration

    def get_report(self):
        return False

    def on_set_report(self, report_data, report_id, report_type):
        # Override this function in order to handle SET REPORT requests from the host,
        # where it sends data to the HID device.
        #
        # This function will only be called if the Report descriptor contains at least one Output entry,
        # and the set_report_buf argument is provided to the constructor.
        #
        # Return True to complete the control transfer normally, False to abort it.
        return True

    def busy(self):
        # Returns True if the interrupt endpoint is busy (i.e. existing transfer is pending)
        return self.is_open() and self.xfer_pending(self._int_ep)

    def send_report(self, report_data, timeout_ms=100):
        # Helper function to send a HID report in the typical USB interrupt
        # endpoint associated with a HID interface.
        #
        # Returns True if successful, False if HID device is not active or timeout
        # is reached without being able to queue the report for sending.
        deadline = time.ticks_add(time.ticks_ms(), timeout_ms)
        while self.busy():
            if time.ticks_diff(deadline, time.ticks_ms()) <= 0:
                return False
            machine.idle()
        if not self.is_open():
            return False
        self.submit_xfer(self._int_ep, report_data)

    def desc_cfg(self, desc, itf_num, ep_num, strs):
        # Add the standard interface descriptor
        desc.interface(
            itf_num,
            1,
            _INTERFACE_CLASS,
            _INTERFACE_SUBCLASS_NONE,
            self.protocol,
            len(strs) if self.interface_str else 0,
        )

        if self.interface_str:
            strs.append(self.interface_str)

        # As per HID v1.11 section 7.1 Standard Requests, return the contents of
        # the standard HID descriptor before the associated endpoint descriptor.
        self.get_hid_descriptor(desc)

        # Add the typical single USB interrupt endpoint descriptor associated
        # with a HID interface.
        self._int_ep = ep_num | _EP_IN_FLAG
        desc.endpoint(self._int_ep, "interrupt", 8, 8)

        self.idle_rate = 0
        self.protocol = 0

    def num_eps(self):
        return 1

    def get_hid_descriptor(self, desc=None):
        # Append a full USB HID descriptor from the object's report descriptor
        # and optional additional descriptors.
        #
        # See HID Specification Version 1.1, Section 6.2.1 HID Descriptor p22

        l = 9 + 3 * len(self.extra_descriptors)  # total length

        if desc is None:
            desc = Descriptor(bytearray(l))

        desc.pack(
            "<BBHBBBH",
            l,  # bLength
            _DESC_HID_TYPE,  # bDescriptorType
            0x111,  # bcdHID
            0,  # bCountryCode
            len(self.extra_descriptors) + 1,  # bNumDescriptors
            0x22,  # bDescriptorType, Report
            len(self.report_descriptor),  # wDescriptorLength, Report
        )
        # Fill in any additional descriptor type/length pairs
        #
        # TODO: unclear if this functionality is ever used, may be easier to not
        # support in base class
        for dt, dd in self.extra_descriptors:
            desc.pack("<BH", dt, len(dd))

        return desc.b

    def on_interface_control_xfer(self, stage, request):
        # Handle standard and class-specific interface control transfers for HID devices.
        bmRequestType, bRequest, wValue, _, wLength = struct.unpack("BBHHH", request)

        recipient, req_type, _ = split_bmRequestType(bmRequestType)

        if stage == _STAGE_SETUP:
            if req_type == _REQ_TYPE_STANDARD:
                # HID Spec p48: 7.1 Standard Requests
                if bRequest == _REQ_CONTROL_GET_DESCRIPTOR:
                    desc_type = wValue >> 8
                    if desc_type == _DESC_HID_TYPE:
                        return self.get_hid_descriptor()
                    if desc_type == _DESC_REPORT_TYPE:
                        return self.report_descriptor
            elif req_type == _REQ_TYPE_CLASS:
                # HID Spec p50: 7.2 Class-Specific Requests
                if bRequest == _REQ_CONTROL_GET_REPORT:
                    print("GET_REPORT?")
                    return False  # Unsupported for now
                if bRequest == _REQ_CONTROL_GET_IDLE:
                    return bytes([self.idle_rate])
                if bRequest == _REQ_CONTROL_GET_PROTOCOL:
                    return bytes([self.protocol])
                if bRequest in (_REQ_CONTROL_SET_IDLE, _REQ_CONTROL_SET_PROTOCOL):
                    return True
                if bRequest == _REQ_CONTROL_SET_REPORT:
                    return self._set_report_buf  # If None, request will stall
            return False  # Unsupported request

        if stage == _STAGE_ACK:
            if req_type == _REQ_TYPE_CLASS:
                if bRequest == _REQ_CONTROL_SET_IDLE:
                    self.idle_rate = wValue >> 8
                elif bRequest == _REQ_CONTROL_SET_PROTOCOL:
                    self.protocol = wValue
                elif bRequest == _REQ_CONTROL_SET_REPORT:
                    report_id = wValue & 0xFF
                    report_type = wValue >> 8
                    report_data = self._set_report_buf
                    if wLength < len(report_data):
                        # need to truncate the response in the callback if we got less bytes
                        # than allowed for in the buffer
                        report_data = memoryview(self._set_report_buf)[:wLength]
                    self.on_set_report(report_data, report_id, report_type)

        return True  # allow DATA/ACK stages to complete normally



"""
Don't use ``tempfile`` in these tests, as ``tempfile`` relies on ``shutil``.
"""

import os
import shutil
import unittest


class TestRmtree(unittest.TestCase):
    def test_dir_dne(self):
        with self.assertRaises(OSError):
            os.stat("foo")

        with self.assertRaises(OSError):
            shutil.rmtree("foo")

    def test_file(self):
        fn = "foo"
        with open(fn, "w"):
            pass

        with self.assertRaises(OSError):
            shutil.rmtree(fn)

        os.remove(fn)

    def test_empty_dir(self):
        with self.assertRaises(OSError):
            # If this triggers, a previous test didn't clean up.
            # bit of a chicken/egg situation with ``tempfile``
            os.stat("foo")

        os.mkdir("foo")
        shutil.rmtree("foo")

        with self.assertRaises(OSError):
            os.stat("foo")

    def test_dir(self):
        with self.assertRaises(OSError):
            # If this triggers, a previous test didn't clean up.
            # bit of a chicken/egg situation with ``tempfile``
            os.stat("foo")

        os.mkdir("foo")
        os.mkdir("foo/bar")
        with open("foo/bar/baz1.txt", "w"):
            pass
        with open("foo/bar/baz2.txt", "w"):
            pass

        shutil.rmtree("foo")

        with self.assertRaises(OSError):
            os.stat("foo")



import sys
import machine
import time

# Verify that machine.idle() resumes execution within 0.1 and 1.1ms (should be
# 1ms max but allowing for some overhead).
#
# (A minimum sleep time for machine.idle() isn't specified but in a system like
# this with no active interrupts then we should expect some idle time before
# resuming. If it's consistently resuming immediately then that indicates a bug
# is preventing proper idle.)
#
# This test doesn't contain any rp2-specific code, but rp2 is currently the only
# tickless port - which is what led to the bug this is a regression test for.
# Some other ports (unix, esp32) have idle behaviour that resumes immediately on
# a quiet system, so this test is also less useful for those.
#
# Verification uses the average idle time, as individual iterations will always
# have outliers due to interrupts, scheduler, etc.

# RP2350 currently fails this test because machine.idle() resumes immediately.
if "RP2350" in sys.implementation._machine:
    print("SKIP")
    raise SystemExit

ITERATIONS = 500
total = 0

for _ in range(ITERATIONS):
    before = time.ticks_us()
    machine.idle()
    total += time.ticks_diff(time.ticks_us(), before)

total /= 1000  # us to ms
average = total / ITERATIONS

# print(f"Total {total}ms average {average}ms")  # uncomment for debug

if 0.1 < average < 1.1:
    print("PASS")
else:
    print(f"Total {total}ms average {average}ms, out of spec")



# Test sys.std* objects.

import sys

try:
    sys.stdout
    sys.stdin
    sys.stderr
except AttributeError:
    print("SKIP")
    raise SystemExit

# CPython is more verbose; no need to match exactly

print('TextIOWrapper' in str(sys.stdout))
print('TextIOWrapper' in str(sys.stderr))
print('TextIOWrapper' in str(sys.stdin))

print('TextIOWrapper' in str(type(sys.stdout)))
print('TextIOWrapper' in str(type(sys.stderr)))
print('TextIOWrapper' in str(type(sys.stdin)))



# test ssl.getpeercert() method

import io
import socket
import ssl

s_test = ssl.wrap_socket(io.BytesIO(), server_side=True, do_handshake=False)
s_test.close()
if not hasattr(s_test, "getpeercert"):
    print("SKIP")
    raise SystemExit


def test(peer_addr):
    s = socket.socket()
    s.connect(peer_addr)
    s = ssl.wrap_socket(s)
    cert = s.getpeercert(True)
    print(type(cert), len(cert) > 100)
    s.close()


if __name__ == "__main__":
    test(socket.getaddrinfo("micropython.org", 443)[0][-1])



# Test Lock class
#
# This test has a corresponding .exp file because it tests a very specific ordering of
# events when cancelling a task waiting on a lock, and that ordering should not change
# (even though it does match CPython's output).

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def task_loop(id, lock):
    print("task start", id)
    for i in range(3):
        async with lock:
            print("task have", id, i)
    print("task end", id)


async def task_sleep(lock):
    async with lock:
        print("task have", lock.locked())
        await asyncio.sleep(0.2)
    print("task release", lock.locked())
    await lock.acquire()
    print("task have again")
    lock.release()


async def task_cancel(id, lock, to_cancel=None):
    try:
        async with lock:
            print("task got", id)
            await asyncio.sleep(0.1)
        print("task release", id)
        if to_cancel:
            to_cancel[0].cancel()
    except asyncio.CancelledError:
        print("task cancel", id)


async def main():
    lock = asyncio.Lock()

    # Basic acquire/release
    print(lock.locked())
    await lock.acquire()
    print(lock.locked())
    await asyncio.sleep(0)
    lock.release()
    print(lock.locked())
    await asyncio.sleep(0)

    # Use with "async with"
    async with lock:
        print("have lock")

    # 3 tasks wanting the lock
    print("----")
    asyncio.create_task(task_loop(1, lock))
    asyncio.create_task(task_loop(2, lock))
    t3 = asyncio.create_task(task_loop(3, lock))
    await lock.acquire()
    await asyncio.sleep(0)
    lock.release()
    await t3

    # 2 sleeping tasks both wanting the lock
    print("----")
    asyncio.create_task(task_sleep(lock))
    await asyncio.sleep(0.1)
    await task_sleep(lock)

    # 3 tasks, the first cancelling the second, the third should still run
    print("----")
    ts = [None]
    asyncio.create_task(task_cancel(0, lock, ts))
    ts[0] = asyncio.create_task(task_cancel(1, lock))
    asyncio.create_task(task_cancel(2, lock))
    await asyncio.sleep(0.3)
    print(lock.locked())

    # 3 tasks, the second and third being cancelled while waiting on the lock
    print("----")
    t0 = asyncio.create_task(task_cancel(0, lock))
    t1 = asyncio.create_task(task_cancel(1, lock))
    t2 = asyncio.create_task(task_cancel(2, lock))
    await asyncio.sleep(0.05)
    t1.cancel()
    await asyncio.sleep(0.1)
    t2.cancel()
    await asyncio.sleep(0.1)
    print(lock.locked())


asyncio.run(main())



# Implements the hmac module from the Python standard library.


class HMAC:
    def __init__(self, key, msg=None, digestmod=None):
        if not isinstance(key, (bytes, bytearray)):
            raise TypeError("key: expected bytes/bytearray")

        import hashlib

        if digestmod is None:
            # TODO: Default hash algorithm is now deprecated.
            digestmod = hashlib.md5

        if callable(digestmod):
            # A hashlib constructor returning a new hash object.
            make_hash = digestmod  # A
        elif isinstance(digestmod, str):
            # A hash name suitable for hashlib.new().
            make_hash = lambda d=b"": getattr(hashlib, digestmod)(d)
        else:
            # A module supporting PEP 247.
            make_hash = digestmod.new  # C

        self._outer = make_hash()
        self._inner = make_hash()

        self.digest_size = getattr(self._inner, "digest_size", None)
        # If the provided hash doesn't support block_size (e.g. built-in
        # hashlib), 64 is the correct default for all built-in hash
        # functions (md5, sha1, sha256).
        self.block_size = getattr(self._inner, "block_size", 64)

        # Truncate to digest_size if greater than block_size.
        if len(key) > self.block_size:
            key = make_hash(key).digest()

        # Pad to block size.
        key = key + bytes(self.block_size - len(key))

        self._outer.update(bytes(x ^ 0x5C for x in key))
        self._inner.update(bytes(x ^ 0x36 for x in key))

        if msg is not None:
            self.update(msg)

    @property
    def name(self):
        return "hmac-" + getattr(self._inner, "name", type(self._inner).__name__)

    def update(self, msg):
        self._inner.update(msg)

    def copy(self):
        if not hasattr(self._inner, "copy"):
            # Not supported for built-in hash functions.
            raise NotImplementedError()
        # Call __new__ directly to avoid the expensive __init__.
        other = self.__class__.__new__(self.__class__)
        other.block_size = self.block_size
        other.digest_size = self.digest_size
        other._inner = self._inner.copy()
        other._outer = self._outer.copy()
        return other

    def _current(self):
        h = self._outer
        if hasattr(h, "copy"):
            # built-in hash functions don't support this, and as a result,
            # digest() will finalise the hmac and further calls to
            # update/digest will fail.
            h = h.copy()
        h.update(self._inner.digest())
        return h

    def digest(self):
        h = self._current()
        return h.digest()

    def hexdigest(self):
        import binascii

        return str(binascii.hexlify(self.digest()), "utf-8")


def new(key, msg=None, digestmod=None):
    return HMAC(key, msg, digestmod)



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2014-2021 Damien P. George
# Copyright (c) 2017 Paul Sokolovsky
# Copyright (c) 2023 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This is based on the serial-only parts of tools/pyboard.py, with Python 2
# support removed, and is currently in the process of being refactored to
# support multiple transports (webrepl, socket, BLE, etc). At the moment,
# SerialTransport is just the old Pyboard+PyboardExtended class without any
# of this refactoring. The API is going to change significantly.

# Once the API is stabilised, the idea is that mpremote can be used both
# as a command line tool and a library for interacting with devices.

import ast, io, os, re, struct, sys, time
from errno import EPERM
from .console import VT_ENABLED
from .transport import TransportError, TransportExecError, Transport


class SerialTransport(Transport):
    def __init__(self, device, baudrate=115200, wait=0, exclusive=True):
        self.in_raw_repl = False
        self.use_raw_paste = True
        self.device_name = device
        self.mounted = False

        import serial
        import serial.tools.list_ports

        # Set options, and exclusive if pyserial supports it
        serial_kwargs = {"baudrate": baudrate, "interCharTimeout": 1}
        if serial.__version__ >= "3.3":
            serial_kwargs["exclusive"] = exclusive

        delayed = False
        for attempt in range(wait + 1):
            try:
                if device.startswith("rfc2217://"):
                    self.serial = serial.serial_for_url(device, **serial_kwargs)
                elif os.name == "nt":
                    self.serial = serial.Serial(**serial_kwargs)
                    self.serial.port = device
                    portinfo = list(serial.tools.list_ports.grep(device))  # type: ignore
                    if portinfo and portinfo[0].manufacturer != "Microsoft":
                        # ESP8266/ESP32 boards use RTS/CTS for flashing and boot mode selection.
                        # DTR False: to avoid using the reset button will hang the MCU in bootloader mode
                        # RTS False: to prevent pulses on rts on serial.close() that would POWERON_RESET an ESPxx
                        self.serial.dtr = False  # DTR False = gpio0 High = Normal boot
                        self.serial.rts = False  # RTS False = EN High = MCU enabled
                    self.serial.open()
                else:
                    self.serial = serial.Serial(device, **serial_kwargs)
                break
            except OSError:
                if wait == 0:
                    continue
                if attempt == 0:
                    sys.stdout.write("Waiting {} seconds for pyboard ".format(wait))
                    delayed = True
            time.sleep(1)
            sys.stdout.write(".")
            sys.stdout.flush()
        else:
            if delayed:
                print("")
            raise TransportError("failed to access " + device)
        if delayed:
            print("")

    def close(self):
        self.serial.close()

    def read_until(self, min_num_bytes, ending, timeout=10, data_consumer=None):
        # if data_consumer is used then data is not accumulated and the ending must be 1 byte long
        assert data_consumer is None or len(ending) == 1

        data = self.serial.read(min_num_bytes)
        if data_consumer:
            data_consumer(data)
        timeout_count = 0
        while True:
            if data.endswith(ending):
                break
            elif self.serial.inWaiting() > 0:
                new_data = self.serial.read(1)
                if data_consumer:
                    data_consumer(new_data)
                    data = new_data
                else:
                    data = data + new_data
                timeout_count = 0
            else:
                timeout_count += 1
                if timeout is not None and timeout_count >= 100 * timeout:
                    break
                time.sleep(0.01)
        return data

    def enter_raw_repl(self, soft_reset=True):
        self.serial.write(b"\r\x03")  # ctrl-C: interrupt any running program

        # flush input (without relying on serial.flushInput())
        n = self.serial.inWaiting()
        while n > 0:
            self.serial.read(n)
            n = self.serial.inWaiting()

        self.serial.write(b"\r\x01")  # ctrl-A: enter raw REPL

        if soft_reset:
            data = self.read_until(1, b"raw REPL; CTRL-B to exit\r\n>")
            if not data.endswith(b"raw REPL; CTRL-B to exit\r\n>"):
                print(data)
                raise TransportError("could not enter raw repl")

            self.serial.write(b"\x04")  # ctrl-D: soft reset

            # Waiting for "soft reboot" independently to "raw REPL" (done below)
            # allows boot.py to print, which will show up after "soft reboot"
            # and before "raw REPL".
            data = self.read_until(1, b"soft reboot\r\n")
            if not data.endswith(b"soft reboot\r\n"):
                print(data)
                raise TransportError("could not enter raw repl")

        data = self.read_until(1, b"raw REPL; CTRL-B to exit\r\n")
        if not data.endswith(b"raw REPL; CTRL-B to exit\r\n"):
            print(data)
            raise TransportError("could not enter raw repl")

        self.in_raw_repl = True

    def exit_raw_repl(self):
        self.serial.write(b"\r\x02")  # ctrl-B: enter friendly REPL
        self.in_raw_repl = False

    def follow(self, timeout, data_consumer=None):
        # wait for normal output
        data = self.read_until(1, b"\x04", timeout=timeout, data_consumer=data_consumer)
        if not data.endswith(b"\x04"):
            raise TransportError("timeout waiting for first EOF reception")
        data = data[:-1]

        # wait for error output
        data_err = self.read_until(1, b"\x04", timeout=timeout)
        if not data_err.endswith(b"\x04"):
            raise TransportError("timeout waiting for second EOF reception")
        data_err = data_err[:-1]

        # return normal and error output
        return data, data_err

    def raw_paste_write(self, command_bytes):
        # Read initial header, with window size.
        data = self.serial.read(2)
        window_size = struct.unpack("<H", data)[0]
        window_remain = window_size

        # Write out the command_bytes data.
        i = 0
        while i < len(command_bytes):
            while window_remain == 0 or self.serial.inWaiting():
                data = self.serial.read(1)
                if data == b"\x01":
                    # Device indicated that a new window of data can be sent.
                    window_remain += window_size
                elif data == b"\x04":
                    # Device indicated abrupt end.  Acknowledge it and finish.
                    self.serial.write(b"\x04")
                    return
                else:
                    # Unexpected data from device.
                    raise TransportError("unexpected read during raw paste: {}".format(data))
            # Send out as much data as possible that fits within the allowed window.
            b = command_bytes[i : min(i + window_remain, len(command_bytes))]
            self.serial.write(b)
            window_remain -= len(b)
            i += len(b)

        # Indicate end of data.
        self.serial.write(b"\x04")

        # Wait for device to acknowledge end of data.
        data = self.read_until(1, b"\x04")
        if not data.endswith(b"\x04"):
            raise TransportError("could not complete raw paste: {}".format(data))

    def exec_raw_no_follow(self, command):
        if isinstance(command, bytes):
            command_bytes = command
        else:
            command_bytes = bytes(command, encoding="utf8")

        # check we have a prompt
        data = self.read_until(1, b">")
        if not data.endswith(b">"):
            raise TransportError("could not enter raw repl")

        if self.use_raw_paste:
            # Try to enter raw-paste mode.
            self.serial.write(b"\x05A\x01")
            data = self.serial.read(2)
            if data == b"R\x00":
                # Device understood raw-paste command but doesn't support it.
                pass
            elif data == b"R\x01":
                # Device supports raw-paste mode, write out the command using this mode.
                return self.raw_paste_write(command_bytes)
            else:
                # Device doesn't support raw-paste, fall back to normal raw REPL.
                data = self.read_until(1, b"w REPL; CTRL-B to exit\r\n>")
                if not data.endswith(b"w REPL; CTRL-B to exit\r\n>"):
                    print(data)
                    raise TransportError("could not enter raw repl")
            # Don't try to use raw-paste mode again for this connection.
            self.use_raw_paste = False

        # Write command using standard raw REPL, 256 bytes every 10ms.
        for i in range(0, len(command_bytes), 256):
            self.serial.write(command_bytes[i : min(i + 256, len(command_bytes))])
            time.sleep(0.01)
        self.serial.write(b"\x04")

        # check if we could exec command
        data = self.serial.read(2)
        if data != b"OK":
            raise TransportError("could not exec command (response: %r)" % data)

    def exec_raw(self, command, timeout=10, data_consumer=None):
        self.exec_raw_no_follow(command)
        return self.follow(timeout, data_consumer)

    def eval(self, expression, parse=True):
        if parse:
            ret = self.exec("print(repr({}))".format(expression))
            ret = ret.strip()
            return ast.literal_eval(ret.decode())
        else:
            ret = self.exec("print({})".format(expression))
            ret = ret.strip()
            return ret

    def exec(self, command, data_consumer=None):
        ret, ret_err = self.exec_raw(command, data_consumer=data_consumer)
        if ret_err:
            raise TransportExecError(ret, ret_err.decode())
        return ret

    def execfile(self, filename):
        with open(filename, "rb") as f:
            pyfile = f.read()
        return self.exec(pyfile)

    def mount_local(self, path, unsafe_links=False):
        fout = self.serial
        if not self.eval('"RemoteFS" in globals()'):
            self.exec(fs_hook_code)
        self.exec("__mount()")
        self.mounted = True
        self.cmd = PyboardCommand(self.serial, fout, path, unsafe_links=unsafe_links)
        self.serial = SerialIntercept(self.serial, self.cmd)

    def write_ctrl_d(self, out_callback):
        self.serial.write(b"\x04")
        if not self.mounted:
            return

        # Read response from the device until it is quiet (with a timeout).
        INITIAL_TIMEOUT = 0.5
        BANNER_TIMEOUT = 2
        QUIET_TIMEOUT = 0.1
        FULL_TIMEOUT = 5
        t_start = t_last_activity = time.monotonic()
        data_all = b""
        soft_reboot_started = False
        soft_reboot_banner = False
        while True:
            t = time.monotonic()
            n = self.serial.inWaiting()
            if n > 0:
                data = self.serial.read(n)
                out_callback(data)
                data_all += data
                t_last_activity = t
            else:
                if len(data_all) == 0:
                    if t - t_start > INITIAL_TIMEOUT:
                        return
                else:
                    if t - t_start > FULL_TIMEOUT:
                        if soft_reboot_started:
                            break
                        return

                    next_data_timeout = QUIET_TIMEOUT

                    if not soft_reboot_started and data_all.find(b"MPY: soft reboot") != -1:
                        soft_reboot_started = True

                    if soft_reboot_started and not soft_reboot_banner:
                        # Once soft reboot has been initiated, give some more time for the startup
                        # banner to be shown
                        if data_all.find(b"\nMicroPython ") != -1:
                            soft_reboot_banner = True
                        elif data_all.find(b"\nraw REPL; CTRL-B to exit\r\n") != -1:
                            soft_reboot_banner = True
                        else:
                            next_data_timeout = BANNER_TIMEOUT

                    if t - t_last_activity > next_data_timeout:
                        break

        if not soft_reboot_started:
            return

        if not soft_reboot_banner:
            out_callback(b"Warning: Could not remount local filesystem\r\n")
            return

        # Determine type of prompt
        if data_all.endswith(b">"):
            in_friendly_repl = False
            prompt = b">"
        else:
            in_friendly_repl = True
            prompt = data_all.rsplit(b"\r\n", 1)[-1]

        # Clear state while board remounts, it will be re-set once mounted.
        self.mounted = False
        self.serial = self.serial.orig_serial

        # Provide a message about the remount.
        out_callback(bytes(f"\r\nRemount local directory {self.cmd.root} at /remote\r\n", "utf8"))

        # Enter raw REPL and re-mount the remote filesystem.
        self.serial.write(b"\x01")
        self.exec(fs_hook_code)
        self.exec("__mount()")
        self.mounted = True

        # Exit raw REPL if needed, and wait for the friendly REPL prompt.
        if in_friendly_repl:
            self.exit_raw_repl()
        self.read_until(len(prompt), prompt)
        out_callback(prompt)
        self.serial = SerialIntercept(self.serial, self.cmd)

    def umount_local(self):
        if self.mounted:
            self.exec('os.umount("/remote")')
            self.mounted = False
            self.serial = self.serial.orig_serial


fs_hook_cmds = {
    "CMD_STAT": 1,
    "CMD_ILISTDIR_START": 2,
    "CMD_ILISTDIR_NEXT": 3,
    "CMD_OPEN": 4,
    "CMD_CLOSE": 5,
    "CMD_READ": 6,
    "CMD_WRITE": 7,
    "CMD_SEEK": 8,
    "CMD_REMOVE": 9,
    "CMD_RENAME": 10,
    "CMD_MKDIR": 11,
    "CMD_RMDIR": 12,
}

fs_hook_code = """\
import os, io, struct, micropython

SEEK_SET = 0

class RemoteCommand:
    def __init__(self):
        import select, sys
        self.buf4 = bytearray(4)
        self.fout = sys.stdout.buffer
        self.fin = sys.stdin.buffer
        self.poller = select.poll()
        self.poller.register(self.fin, select.POLLIN)

    def poll_in(self):
        for _ in self.poller.ipoll(1000):
            return
        self.end()
        raise Exception('timeout waiting for remote')

    def rd(self, n):
        buf = bytearray(n)
        self.rd_into(buf, n)
        return buf

    def rd_into(self, buf, n):
        # implement reading with a timeout in case other side disappears
        if n == 0:
            return
        self.poll_in()
        r = self.fin.readinto(buf, n)
        if r < n:
            mv = memoryview(buf)
            while r < n:
                self.poll_in()
                r += self.fin.readinto(mv[r:], n - r)

    def begin(self, type):
        micropython.kbd_intr(-1)
        buf4 = self.buf4
        buf4[0] = 0x18
        buf4[1] = type
        self.fout.write(buf4, 2)
        # Wait for sync byte 0x18, but don't get stuck forever
        for i in range(30):
            self.poller.poll(1000)
            self.fin.readinto(buf4, 1)
            if buf4[0] == 0x18:
                break

    def end(self):
        micropython.kbd_intr(3)

    def rd_s8(self):
        self.rd_into(self.buf4, 1)
        n = self.buf4[0]
        if n & 0x80:
            n -= 0x100
        return n

    def rd_s32(self):
        buf4 = self.buf4
        self.rd_into(buf4, 4)
        n = buf4[0] | buf4[1] << 8 | buf4[2] << 16 | buf4[3] << 24
        if buf4[3] & 0x80:
            n -= 0x100000000
        return n

    def rd_u32(self):
        buf4 = self.buf4
        self.rd_into(buf4, 4)
        return buf4[0] | buf4[1] << 8 | buf4[2] << 16 | buf4[3] << 24

    def rd_bytes(self, buf):
        # TODO if n is large (eg >256) then we may miss bytes on stdin
        n = self.rd_s32()
        if buf is None:
            ret = buf = bytearray(n)
        else:
            ret = n
        self.rd_into(buf, n)
        return ret

    def rd_str(self):
        n = self.rd_s32()
        if n == 0:
            return ''
        else:
            return str(self.rd(n), 'utf8')

    def wr_s8(self, i):
        self.buf4[0] = i
        self.fout.write(self.buf4, 1)

    def wr_s32(self, i):
        struct.pack_into('<i', self.buf4, 0, i)
        self.fout.write(self.buf4)

    def wr_bytes(self, b):
        self.wr_s32(len(b))
        self.fout.write(b)

    # str and bytes act the same in MicroPython
    wr_str = wr_bytes


class RemoteFile(io.IOBase):
    def __init__(self, cmd, fd, is_text):
        self.cmd = cmd
        self.fd = fd
        self.is_text = is_text

    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        l = self.readline()
        if not l:
            raise StopIteration
        return l

    def ioctl(self, request, arg):
        if request == 1:  # FLUSH
            self.flush()
        elif request == 2:  # SEEK
            # This assumes a 32-bit bare-metal machine.
            import machine
            machine.mem32[arg] = self.seek(machine.mem32[arg], machine.mem32[arg + 4])
        elif request == 4:  # CLOSE
            self.close()
        elif request == 11:  # BUFFER_SIZE
            # This is used as the vfs_reader buffer. n + 4 should be less than 255 to
            # fit in stdin ringbuffer on supported ports. n + 7 should be multiple of 16
            # to efficiently use gc blocks in mp_reader_vfs_t.
            return 249
        else:
            return -1
        return 0

    def flush(self):
        pass

    def close(self):
        if self.fd is None:
            return
        c = self.cmd
        c.begin(CMD_CLOSE)
        c.wr_s8(self.fd)
        c.end()
        self.fd = None

    def read(self, n=-1):
        c = self.cmd
        c.begin(CMD_READ)
        c.wr_s8(self.fd)
        c.wr_s32(n)
        data = c.rd_bytes(None)
        c.end()
        if self.is_text:
            data = str(data, 'utf8')
        else:
            data = bytes(data)
        return data

    def readinto(self, buf):
        c = self.cmd
        c.begin(CMD_READ)
        c.wr_s8(self.fd)
        c.wr_s32(len(buf))
        n = c.rd_bytes(buf)
        c.end()
        return n

    def readline(self):
        l = ''
        while 1:
            c = self.read(1)
            l += c
            if c == '\\n' or c == '':
                return l

    def readlines(self):
        ls = []
        while 1:
            l = self.readline()
            if not l:
                return ls
            ls.append(l)

    def write(self, buf):
        c = self.cmd
        c.begin(CMD_WRITE)
        c.wr_s8(self.fd)
        c.wr_bytes(buf)
        n = c.rd_s32()
        c.end()
        return n

    def seek(self, n, whence=SEEK_SET):
        c = self.cmd
        c.begin(CMD_SEEK)
        c.wr_s8(self.fd)
        c.wr_s32(n)
        c.wr_s8(whence)
        n = c.rd_s32()
        c.end()
        if n < 0:
            raise OSError(n)
        return n


class RemoteFS:
    def __init__(self, cmd):
        self.cmd = cmd

    def _abspath(self, path):
        return path if path.startswith("/") else self.path + path

    def mount(self, readonly, mkfs):
        pass

    def umount(self):
        pass

    def chdir(self, path):
        if not path.startswith("/"):
            path = self.path + path
        if not path.endswith("/"):
            path += "/"
        if path != "/":
            self.stat(path)
        self.path = path

    def getcwd(self):
        return self.path

    def remove(self, path):
        c = self.cmd
        c.begin(CMD_REMOVE)
        c.wr_str(self._abspath(path))
        res = c.rd_s32()
        c.end()
        if res < 0:
            raise OSError(-res)

    def rename(self, old, new):
        c = self.cmd
        c.begin(CMD_RENAME)
        c.wr_str(self._abspath(old))
        c.wr_str(self._abspath(new))
        res = c.rd_s32()
        c.end()
        if res < 0:
            raise OSError(-res)

    def mkdir(self, path):
        c = self.cmd
        c.begin(CMD_MKDIR)
        c.wr_str(self._abspath(path))
        res = c.rd_s32()
        c.end()
        if res < 0:
            raise OSError(-res)

    def rmdir(self, path):
        c = self.cmd
        c.begin(CMD_RMDIR)
        c.wr_str(self._abspath(path))
        res = c.rd_s32()
        c.end()
        if res < 0:
            raise OSError(-res)

    def stat(self, path):
        c = self.cmd
        c.begin(CMD_STAT)
        c.wr_str(self._abspath(path))
        res = c.rd_s8()
        if res < 0:
            c.end()
            raise OSError(-res)
        mode = c.rd_u32()
        size = c.rd_u32()
        atime = c.rd_u32()
        mtime = c.rd_u32()
        ctime = c.rd_u32()
        c.end()
        return mode, 0, 0, 0, 0, 0, size, atime, mtime, ctime

    def ilistdir(self, path):
        c = self.cmd
        c.begin(CMD_ILISTDIR_START)
        c.wr_str(self._abspath(path))
        res = c.rd_s8()
        c.end()
        if res < 0:
            raise OSError(-res)
        def next():
            while True:
                c.begin(CMD_ILISTDIR_NEXT)
                name = c.rd_str()
                if name:
                    type = c.rd_u32()
                    c.end()
                    yield (name, type, 0)
                else:
                    c.end()
                    break
        return next()

    def open(self, path, mode):
        c = self.cmd
        c.begin(CMD_OPEN)
        c.wr_str(self._abspath(path))
        c.wr_str(mode)
        fd = c.rd_s8()
        c.end()
        if fd < 0:
            raise OSError(-fd)
        return RemoteFile(c, fd, mode.find('b') == -1)


def __mount():
    os.mount(RemoteFS(RemoteCommand()), '/remote')
    os.chdir('/remote')
"""

# Apply basic compression on hook code.
for key, value in fs_hook_cmds.items():
    fs_hook_code = re.sub(key, str(value), fs_hook_code)
fs_hook_code = re.sub(" *#.*$", "", fs_hook_code, flags=re.MULTILINE)
fs_hook_code = re.sub("\n\n+", "\n", fs_hook_code)
fs_hook_code = re.sub("    ", " ", fs_hook_code)
fs_hook_code = re.sub("rd_", "r", fs_hook_code)
fs_hook_code = re.sub("wr_", "w", fs_hook_code)
fs_hook_code = re.sub("buf4", "b4", fs_hook_code)


class PyboardCommand:
    def __init__(self, fin, fout, path, unsafe_links=False):
        self.fin = fin
        self.fout = fout
        self.root = path + "/"
        self.data_ilistdir = ["", []]
        self.data_files = []
        self.unsafe_links = unsafe_links

    def rd_s8(self):
        return struct.unpack("<b", self.fin.read(1))[0]

    def rd_s32(self):
        return struct.unpack("<i", self.fin.read(4))[0]

    def rd_bytes(self):
        n = self.rd_s32()
        return self.fin.read(n)

    def rd_str(self):
        n = self.rd_s32()
        if n == 0:
            return ""
        else:
            return str(self.fin.read(n), "utf8")

    def wr_s8(self, i):
        self.fout.write(struct.pack("<b", i))

    def wr_s32(self, i):
        self.fout.write(struct.pack("<i", i))

    def wr_u32(self, i):
        self.fout.write(struct.pack("<I", i))

    def wr_bytes(self, b):
        self.wr_s32(len(b))
        self.fout.write(b)

    def wr_str(self, s):
        b = bytes(s, "utf8")
        self.wr_s32(len(b))
        self.fout.write(b)

    def log_cmd(self, msg):
        print(f"[{msg}]", end="\r\n")

    def path_check(self, path):
        if not self.unsafe_links:
            parent = os.path.realpath(self.root)
            child = os.path.realpath(path)
        else:
            parent = os.path.abspath(self.root)
            child = os.path.abspath(path)
        if parent != os.path.commonpath([parent, child]):
            raise OSError(EPERM, "")  # File is outside mounted dir

    def do_stat(self):
        path = self.root + self.rd_str()
        # self.log_cmd(f"stat {path}")
        try:
            self.path_check(path)
            stat = os.stat(path)
        except OSError as er:
            self.wr_s8(-abs(er.errno))
        else:
            self.wr_s8(0)
            # Note: st_ino would need to be 64-bit if added here
            self.wr_u32(stat.st_mode)
            self.wr_u32(stat.st_size)
            self.wr_u32(int(stat.st_atime))
            self.wr_u32(int(stat.st_mtime))
            self.wr_u32(int(stat.st_ctime))

    def do_ilistdir_start(self):
        path = self.root + self.rd_str()
        try:
            self.path_check(path)
            self.data_ilistdir[0] = path
            self.data_ilistdir[1] = os.listdir(path)
            self.wr_s8(0)
        except OSError as er:
            self.wr_s8(-abs(er.errno))

    def do_ilistdir_next(self):
        if self.data_ilistdir[1]:
            entry = self.data_ilistdir[1].pop(0)
            try:
                stat = os.lstat(self.data_ilistdir[0] + "/" + entry)
                mode = stat.st_mode & 0xC000
            except OSError:
                mode = 0
            self.wr_str(entry)
            self.wr_u32(mode)
        else:
            self.wr_str("")

    def do_open(self):
        path = self.root + self.rd_str()
        mode = self.rd_str()
        # self.log_cmd(f"open {path} {mode}")
        try:
            self.path_check(path)
            f = open(path, mode)
        except OSError as er:
            self.wr_s8(-abs(er.errno))
        else:
            is_text = mode.find("b") == -1
            try:
                fd = self.data_files.index(None)
                self.data_files[fd] = (f, is_text)
            except ValueError:
                fd = len(self.data_files)
                self.data_files.append((f, is_text))
            self.wr_s8(fd)

    def do_close(self):
        fd = self.rd_s8()
        # self.log_cmd(f"close {fd}")
        self.data_files[fd][0].close()
        self.data_files[fd] = None

    def do_read(self):
        fd = self.rd_s8()
        n = self.rd_s32()
        buf = self.data_files[fd][0].read(n)
        if self.data_files[fd][1]:
            buf = bytes(buf, "utf8")
        self.wr_bytes(buf)
        # self.log_cmd(f"read {fd} {n} -> {len(buf)}")

    def do_seek(self):
        fd = self.rd_s8()
        n = self.rd_s32()
        whence = self.rd_s8()
        # self.log_cmd(f"seek {fd} {n}")
        try:
            n = self.data_files[fd][0].seek(n, whence)
        except io.UnsupportedOperation:
            n = -1
        self.wr_s32(n)

    def do_write(self):
        fd = self.rd_s8()
        buf = self.rd_bytes()
        if self.data_files[fd][1]:
            buf = str(buf, "utf8")
        n = self.data_files[fd][0].write(buf)
        self.wr_s32(n)
        # self.log_cmd(f"write {fd} {len(buf)} -> {n}")

    def do_remove(self):
        path = self.root + self.rd_str()
        # self.log_cmd(f"remove {path}")
        try:
            self.path_check(path)
            os.remove(path)
            ret = 0
        except OSError as er:
            ret = -abs(er.errno)
        self.wr_s32(ret)

    def do_rename(self):
        old = self.root + self.rd_str()
        new = self.root + self.rd_str()
        # self.log_cmd(f"rename {old} {new}")
        try:
            self.path_check(old)
            self.path_check(new)
            os.rename(old, new)
            ret = 0
        except OSError as er:
            ret = -abs(er.errno)
        self.wr_s32(ret)

    def do_mkdir(self):
        path = self.root + self.rd_str()
        # self.log_cmd(f"mkdir {path}")
        try:
            self.path_check(path)
            os.mkdir(path)
            ret = 0
        except OSError as er:
            ret = -abs(er.errno)
        self.wr_s32(ret)

    def do_rmdir(self):
        path = self.root + self.rd_str()
        # self.log_cmd(f"rmdir {path}")
        try:
            self.path_check(path)
            os.rmdir(path)
            ret = 0
        except OSError as er:
            ret = -abs(er.errno)
        self.wr_s32(ret)

    cmd_table = {
        fs_hook_cmds["CMD_STAT"]: do_stat,
        fs_hook_cmds["CMD_ILISTDIR_START"]: do_ilistdir_start,
        fs_hook_cmds["CMD_ILISTDIR_NEXT"]: do_ilistdir_next,
        fs_hook_cmds["CMD_OPEN"]: do_open,
        fs_hook_cmds["CMD_CLOSE"]: do_close,
        fs_hook_cmds["CMD_READ"]: do_read,
        fs_hook_cmds["CMD_WRITE"]: do_write,
        fs_hook_cmds["CMD_SEEK"]: do_seek,
        fs_hook_cmds["CMD_REMOVE"]: do_remove,
        fs_hook_cmds["CMD_RENAME"]: do_rename,
        fs_hook_cmds["CMD_MKDIR"]: do_mkdir,
        fs_hook_cmds["CMD_RMDIR"]: do_rmdir,
    }


class SerialIntercept:
    def __init__(self, serial, cmd):
        self.orig_serial = serial
        self.cmd = cmd
        self.buf = b""
        self.orig_serial.timeout = 5.0

    def _check_input(self, blocking):
        if blocking or self.orig_serial.inWaiting() > 0:
            c = self.orig_serial.read(1)
            if c == b"\x18":
                # a special command
                c = self.orig_serial.read(1)[0]
                self.orig_serial.write(b"\x18")  # Acknowledge command
                PyboardCommand.cmd_table[c](self.cmd)
            elif not VT_ENABLED and c == b"\x1b":
                # ESC code, ignore these on windows
                esctype = self.orig_serial.read(1)
                if esctype == b"[":  # CSI
                    while not (0x40 < self.orig_serial.read(1)[0] < 0x7E):
                        # Looking for "final byte" of escape sequence
                        pass
            else:
                self.buf += c

    @property
    def fd(self):
        return self.orig_serial.fd

    def close(self):
        self.orig_serial.close()

    def inWaiting(self):
        self._check_input(False)
        return len(self.buf)

    def read(self, n):
        while len(self.buf) < n:
            self._check_input(True)
        out = self.buf[:n]
        self.buf = self.buf[n:]
        return out

    def write(self, buf):
        self.orig_serial.write(buf)



import gc
import vfs
import samd
import sys

bdev = samd.Flash()

# Try to mount the filesystem, and format the flash if it doesn't exist.
fs_type = vfs.VfsLfs2 if hasattr(vfs, "VfsLfs2") else vfs.VfsLfs1

try:
    fs = fs_type(bdev, progsize=256)
except:
    fs_type.mkfs(bdev, progsize=256)
    fs = fs_type(bdev, progsize=256)
vfs.mount(fs, "/")
sys.path.append("/lib")

del fs, fs_type, bdev, vfs, samd, sys
gc.collect()
del gc



"""
categories: Types,memoryview
description: memoryview can become invalid if its target is resized
cause: CPython prevents a ``bytearray`` or ``io.bytesIO`` object from changing size while there is a ``memoryview`` object that references it. MicroPython requires the programmer to manually ensure that an object is not resized while any ``memoryview`` references it.

In the worst case scenario, resizing an object which is the target of a memoryview can cause the memoryview(s) to reference invalid freed memory (a use-after-free bug) and corrupt the MicroPython runtime.
workaround: Do not change the size of any ``bytearray`` or ``io.bytesIO`` object that has a ``memoryview`` assigned to it.
"""
b = bytearray(b"abcdefg")
m = memoryview(b)
b.extend(b"hijklmnop")
print(b, bytes(m))



# test select.poll on UDP sockets

try:
    import socket, select

    select.poll  # Raises AttributeError for CPython implementations without poll()
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit

try:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.bind(socket.getaddrinfo("127.0.0.1", 8000)[0][-1])
except OSError:
    print("SKIP")
    raise SystemExit

poll = select.poll()

# UDP socket should not be readable
poll.register(s, select.POLLIN)
print(len(poll.poll(0)))

# UDP socket should be writable
poll.modify(s, select.POLLOUT)
print(poll.poll(0)[0][1] == select.POLLOUT)

# same test for select.select, but just skip it if the function isn't available
if hasattr(select, "select"):
    r, w, e = select.select([s], [], [], 0)
    assert not r and not w and not e



module("_boot.py", opt=3)
module("apa102.py", base_path="$(PORT_DIR)/modules", opt=3)
module("port_diag.py", base_path="$(PORT_DIR)/modules", opt=3)
require("ntptime")
require("dht")
require("onewire")
require("ds18x20")
require("webrepl")
require("neopixel")



# test sys module

import sys
print(sys.__name__)
print(type(sys.path))
print(type(sys.argv))
print(sys.byteorder in ('little', 'big'))

try:
    print(sys.maxsize > 100)
except AttributeError:
    # Effectively skip subtests
    print(True)

try:
    print(sys.implementation.name in ('cpython', 'micropython'))
except AttributeError:
    # Effectively skip subtests
    print(True)

if hasattr(sys.implementation, '_mpy'):
    print(type(sys.implementation._mpy))
else:
    # Effectively skip subtests
    print(int)

try:
    print(sys.intern('micropython') == 'micropython')
    has_intern = True
except AttributeError:
    has_intern = False
    print(True)

if has_intern:
    try:
        print(sys.intern(0))
    except TypeError:
        print(True)
else:
    print(True)



# test re.sub with unmatched groups, behaviour changed in CPython 3.5

try:
    import re
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    re.sub
except AttributeError:
    print("SKIP")
    raise SystemExit

# first group matches, second optional group doesn't so is replaced with a blank
print(re.sub(r"(a)(b)?", r"\2-\1", "1a2"))



# Test the esp32 NVS class - access to esp-idf's Non-Volatile-Storage

from esp32 import NVS

nvs = NVS("mp-test")

# test setting and gettin an integer kv
nvs.set_i32("key1", 1234)
print(nvs.get_i32("key1"))
nvs.set_i32("key2", -503)
print(nvs.get_i32("key2"))
print(nvs.get_i32("key1"))

# test setting and getting a blob kv using a bytearray
blob1 = "testing a string as a blob"
nvs.set_blob("blob1", blob1)
buf1 = bytearray(len(blob1))
len1 = nvs.get_blob("blob1", buf1)
print(buf1)
print(len(blob1), len1)

# test setting and getting a blob kv using a string
blob2 = b"testing a bytearray"
nvs.set_blob("blob2", blob2)
buf2 = bytearray(len(blob2))
len2 = nvs.get_blob("blob2", buf2)
print(buf2)
print(len(blob2), len2)

# test raising of error exceptions
nvs.erase_key("key1")
try:
    nvs.erase_key("key1")  # not found
except OSError as e:
    print(e)
try:
    nvs.get_i32("key1")  # not found
except OSError as e:
    print(e)
try:
    nvs.get_i32("blob1")  # not found (blob1 exists but diff type)
except OSError as e:
    print(e)
try:
    buf3 = bytearray(10)
    nvs.get_blob("blob1", buf3)  # invalid length (too short)
except OSError as e:
    print(e)

nvs.commit()  # we're not verifying that this does anything, just doesn't error

# test using a second namespace and that it doesn't interfere with first
nvs2 = NVS("mp-test2")
try:
    print(nvs2.get_i32("key2"))
except OSError as e:
    print(e)
nvs2.set_i32("key2", 7654)
print(nvs.get_i32("key2"))
print(nvs2.get_i32("key2"))

# clean-up (the namespaces will remain)
nvs.erase_key("key2")
nvs.erase_key("blob1")
nvs.erase_key("blob2")
nvs2.erase_key("key2")
nvs.commit()



# LOLIN S2 MINI MicroPython Helper Library

from micropython import const
from machine import Pin

# Pin Assignments

# SPI
SPI_MOSI = const(11)
SPI_MISO = const(9)
SPI_CLK = const(7)

# I2C
I2C_SDA = const(33)
I2C_SCL = const(35)

# DAC
DAC1 = const(17)
DAC2 = const(18)

# LED
LED = const(15)

# BUTTON
BUTTON = const(0)

# Helper methods for built in sensors

led = Pin(LED, Pin.OUT, value=0)

button = Pin(BUTTON, Pin.IN, Pin.PULL_UP)



# Test that integers format to exact values.

for b in [13, 123, 457, 23456]:
    for r in range(1, 10):
        e_fmt = "{:." + str(r) + "e}"
        f_fmt = "{:." + str(r) + "f}"
        g_fmt = "{:." + str(r) + "g}"
        for e in range(0, 5):
            f = b * (10**e)
            title = str(b) + " x 10^" + str(e)
            print(title, "with format", e_fmt, "gives", e_fmt.format(f))
            print(title, "with format", f_fmt, "gives", f_fmt.format(f))
            print(title, "with format", g_fmt, "gives", g_fmt.format(f))

# 16777215 is 2^24 - 1, the largest integer that can be completely held
# in a float32.
print("{:f}".format(16777215))
# 4294967040 = 16777215 * 128 is the largest integer that is exactly
# represented by a float32 and that will also fit within a (signed) int32.
# The upper bound of our integer-handling code is actually double this,
# but that constant might cause trouble on systems using 32 bit ints.
print("{:f}".format(2147483520))
# Very large positive integers can be a test for precision and resolution.
# This is a weird way to represent 1e38 (largest power of 10 for float32).
print("{:.6e}".format(float("9" * 30 + "e8")))



"""LILYGO TTGO LoRa32 MicroPython Helper Library."""

from machine import Pin, SoftI2C, Signal

from lilygo_oled import OLED

from micropython import const


class Lora32Base:
    """Base class defining common pins."""

    def __init__(self, define_helpers=True):
        # LORA
        self.LORA_MOSI = const(27)
        self.LORA_MISO = const(19)
        self.LORA_SCLK = const(5)
        self.LORA_CS = const(18)
        self.LORA_DIO = const(26)
        self.LORA_RST = const(23)

        # DAC
        self.DAC1 = const(26)

        # LED
        self.LED = const(25)

        # OLED
        self.OLED_SDA = const(21)
        self.OLED_SCL = const(22)
        self.OLED_RST = None

        if define_helpers:
            self.create_helpers()

    def create_helpers(self):
        self.led = Pin(self.LED, Pin.OUT)
        self.i2c = SoftI2C(scl=Pin(self.OLED_SCL), sda=Pin(self.OLED_SDA))
        rstpin = self.OLED_RST is not None and Pin(self.OLED_RST, Pin.OUT) or None
        self.oled = OLED(self.i2c, rstpin)


class Lora32v1_0(Lora32Base):
    """Device Support for LILYGO TTGO LoRa32 v1.0."""

    def __init__(self):
        super().__init__(define_helpers=False)

        # v1.0 has different pins for the following
        self.LORA_RST = const(14)
        self.OLED_SDA = const(4)
        self.OLED_SCL = const(15)

        # Also has a reset for the OLED that the others don't have
        self.OLED_RST = const(16)

        super().create_helpers()


class Lora32v1_2(Lora32Base):
    """Device Support for LILYGO TTGO LoRa32 v1.2 (T-Fox)."""

    def __init__(self):
        super().__init__()

        # v1.2 Has a DS3231 RTC
        self.DS3231_SDA = const(21)
        self.DS3231_SCL = const(22)


class Lora32(Lora32Base):
    """Device Support for LILYGO TTGO LoRa32 v1.6 and v2.0."""

    def __init__(self):
        super().__init__()

        # v1.6 and v2.0 support an SDCard
        self.SD_CS = const(13)
        self.SD_MOSI = const(15)
        self.SD_MISO = const(2)
        self.SD_SCLK = const(14)



# Example using PWM to fade an LED.
# Note: this does not work on Pico W because it uses Pin(25) for LED output.

# ruff: noqa: F821 - @asm_pio decorator adds names to function scope

import time
from machine import Pin, PWM


# Construct PWM object, with LED on Pin(25).
pwm = PWM(Pin(25))

# Set the PWM frequency.
pwm.freq(1000)

# Fade the LED in and out a few times.
duty = 0
direction = 1
for _ in range(8 * 256):
    duty += direction
    if duty > 255:
        duty = 255
        direction = -1
    elif duty < 0:
        duty = 0
        direction = 1
    pwm.duty_u16(duty * duty)
    time.sleep(0.001)



# test bignum comparisons

i = 1 << 65
cases = (0, 1, -1, i, -i, i + 1, -(i + 1))

for lhs in cases:
    for rhs in cases:
        print("{} == {} = {}".format(lhs, rhs, lhs == rhs))
        print("{} != {} = {}".format(lhs, rhs, lhs != rhs))
        print("{} < {} = {}".format(lhs, rhs, lhs < rhs))
        print("{} > {} = {}".format(lhs, rhs, lhs > rhs))
        print("{} <= {} = {}".format(lhs, rhs, lhs <= rhs))
        print("{} >= {} = {}".format(lhs, rhs, lhs >= rhs))



# MicroPython lora simple_rxtx example - synchronous API version
# MIT license; Copyright (c) 2023 Angus Gratton
import time
from machine import Pin, SPI


def get_modem():
    # from lora import SX1276
    #
    # lora_cfg = {
    #    "freq_khz": 916000,
    #    "sf": 8,
    #    "bw": "500",  # kHz
    #    "coding_rate": 8,
    #    "preamble_len": 12,
    #    "output_power": 0,  # dBm
    # }
    #
    # return SX1276(
    #     spi=SPI(1, baudrate=2000_000, polarity=0, phase=0,
    #             miso=Pin(19), mosi=Pin(27), sck=Pin(5)),
    #     cs=Pin(18),
    #     dio0=Pin(26),
    #     dio1=Pin(35),
    #     reset=Pin(14),
    #     lora_cfg=lora_cfg,
    # )
    raise NotImplementedError("Replace this function with one that returns a lora modem instance")


def main():
    print("Initializing...")
    modem = get_modem()

    counter = 0
    while True:
        print("Sending...")
        modem.send(f"Hello world from MicroPython #{counter}".encode())

        print("Receiving...")
        rx = modem.recv(timeout_ms=5000)
        if rx:
            print(f"Received: {rx!r}")
        else:
            print("Timeout!")
        time.sleep(2)
        counter += 1


if __name__ == "__main__":
    main()



# DHT11/DHT22 driver for MicroPython on ESP8266
# MIT license; Copyright (c) 2016 Damien P. George

import sys
import machine

if hasattr(machine, "dht_readinto"):
    from machine import dht_readinto
elif sys.platform.startswith("esp"):
    from esp import dht_readinto
elif sys.platform == "pyboard":
    from pyb import dht_readinto
else:
    dht_readinto = __import__(sys.platform).dht_readinto

del machine


class DHTBase:
    def __init__(self, pin):
        self.pin = pin
        self.buf = bytearray(5)

    def measure(self):
        buf = self.buf
        dht_readinto(self.pin, buf)
        if (buf[0] + buf[1] + buf[2] + buf[3]) & 0xFF != buf[4]:
            raise Exception("checksum error")


class DHT11(DHTBase):
    def humidity(self):
        return self.buf[0]

    def temperature(self):
        return self.buf[2]


class DHT22(DHTBase):
    def humidity(self):
        return (self.buf[0] << 8 | self.buf[1]) * 0.1

    def temperature(self):
        t = ((self.buf[2] & 0x7F) << 8 | self.buf[3]) * 0.1
        if self.buf[2] & 0x80:
            t = -t
        return t



# This example shows how to access Video BIOS memory area via machine.mem
# It requires root privilege and x86 legacy hardware (which has mentioned
# Video BIOS at all).
# It is expected to print 0xaa55, which is a signature at the start of
# Video BIOS.

import machine

print(hex(machine.mem16[0xC0000]))



include("$(PORT_DIR)/variants/manifest.py")

require("abc")
require("base64")
require("collections")
require("collections-defaultdict")
require("copy")
require("datetime")
require("fnmatch")
require("functools")
require("gzip")
require("hmac")
require("html")
require("inspect")
require("io")
require("itertools")
require("locale")
require("logging")
require("operator")
require("os")
require("os-path")
require("pathlib")
require("stat")
require("tarfile")
require("tarfile-write")
require("time")
require("unittest")
require("uu")
require("zlib")
